# Modified by CogDL Team
#
# DeepSpeed note, code taken from commit 3d59216cec89a363649b4fe3d15295ba936ced0f
# https://github.com/NVIDIA/DeepLearningExamples/blob/master/PyTorch/LanguageModeling/BERT/modeling.py

# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors and The HugginFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PyTorch BERT model."""

from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import json
import logging
import math

import torch
from torch import nn
from torch.utils import checkpoint

from torch.nn import Module
from torch.nn.parameter import Parameter
import torch.nn.functional as F
import torch.nn.init as init


logger = logging.getLogger(__name__)


@torch.jit.script
def f_gelu(x):
    pdtype = x.dtype
    x = x.float()
    y = x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))
    return y.to(pdtype)


@torch.jit.script
def bias_gelu(bias, y):
    x = bias + y
    return x * 0.5 * (1.0 + torch.erf(x / 1.41421))


@torch.jit.script
def bias_tanh(bias, y):
    x = bias + y
    return torch.tanh(x)


def gelu(x):
    """Implementation of the gelu activation function.
    For information: OpenAI GPT's gelu is slightly different (and gives slightly different results):
    0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))
    Also see https://arxiv.org/abs/1606.08415
    """
    return f_gelu(x)


def swish(x):
    return x * torch.sigmoid(x)


ACT2FN = {"gelu": gelu, "relu": torch.nn.functional.relu, "swish": swish}


class LinearActivation(Module):
    r"""Fused Linear and activation Module."""
    __constants__ = ["bias"]

    def __init__(self, in_features, out_features, act="gelu", bias=True):
        super(LinearActivation, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.fused_gelu = False
        self.fused_tanh = False
        if isinstance(act, str):
            if bias and act == "gelu":
                self.fused_gelu = True
            elif bias and act == "tanh":
                self.fused_tanh = True
            else:
                self.act_fn = ACT2FN[act]
        else:
            self.act_fn = act
        self.weight = Parameter(torch.Tensor(out_features, in_features))
        if bias:
            self.bias = Parameter(torch.Tensor(out_features))
        else:
            self.register_parameter("bias", None)
        self.reset_parameters()

    def reset_parameters(self):
        init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in)
            init.uniform_(self.bias, -bound, bound)

    def forward(self, input):
        if self.fused_gelu:
            return bias_gelu(self.bias, F.linear(input, self.weight, None))
        elif self.fused_tanh:
            return bias_tanh(self.bias, F.linear(input, self.weight, None))
        else:
            return self.act_fn(F.linear(input, self.weight, self.bias))

    def extra_repr(self):
        return "in_features={}, out_features={}, bias={}".format(
            self.in_features, self.out_features, self.bias is not None
        )


class BertConfig(object):
    """Configuration class to store the configuration of a `BertModel`."""

    def __init__(self):
        pass

    @classmethod
    def from_dict(cls, json_object):
        """Constructs a `BertConfig` from a Python dictionary of parameters."""
        config = cls()
        for key, value in json_object.items():
            config.__dict__[key] = value
        return config

    def __repr__(self):
        return str(self.to_json_string())

    def to_dict(self):
        """Serializes this instance to a Python dictionary."""
        output = copy.deepcopy(self.__dict__)
        return output

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"


class BertLayerNorm(nn.Module):
    def __init__(self, hidden_size, eps=1e-12):
        """Construct a layernorm module in the TF style (epsilon inside the square root)."""
        super(BertLayerNorm, self).__init__()
        self.weight = nn.Parameter(torch.ones(hidden_size))
        self.bias = nn.Parameter(torch.zeros(hidden_size))
        self.variance_epsilon = eps

    def forward(self, x):
        pdtype = x.dtype
        x = x.float()
        u = x.mean(-1, keepdim=True)
        s = (x - u).pow(2).mean(-1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.variance_epsilon)
        return self.weight * x.to(pdtype) + self.bias


class BertEmbeddings(nn.Module):
    """Construct the embeddings from word, position and token_type embeddings."""

    def __init__(self, config):
        super(BertEmbeddings, self).__init__()
        self.word_embeddings = nn.Embedding(config.vocab_size, config.hidden_size)
        self.position_embeddings = nn.Embedding(config.max_position_embeddings, config.hidden_size)
        self.token_type_embeddings = nn.Embedding(config.type_vocab_size, config.hidden_size)
        # self.LayerNorm is not snake-cased to stick with TensorFlow model variable name and be able to load
        # any TensorFlow checkpoint file
        self.LayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, input_ids, token_type_ids=None):
        seq_length = input_ids.size(1)
        position_ids = torch.arange(seq_length, dtype=torch.long, device=input_ids.device)
        position_ids = position_ids.unsqueeze(0).expand_as(input_ids)
        if token_type_ids is None:
            token_type_ids = torch.zeros_like(input_ids)

        words_embeddings = self.word_embeddings(input_ids)
        position_embeddings = self.position_embeddings(position_ids)
        token_type_embeddings = self.token_type_embeddings(token_type_ids)

        embeddings = words_embeddings + position_embeddings + token_type_embeddings
        embeddings = self.LayerNorm(embeddings)
        embeddings = self.dropout(embeddings)
        return embeddings


class BertSelfAttention(nn.Module):
    def __init__(self, config):
        super(BertSelfAttention, self).__init__()
        if config.hidden_size % config.num_attention_heads != 0:
            raise ValueError(
                "The hidden size (%d) is not a multiple of the number of attention "
                "heads (%d)" % (config.hidden_size, config.num_attention_heads)
            )
        self.num_attention_heads = config.num_attention_heads
        self.attention_head_size = int(config.hidden_size / config.num_attention_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = nn.Linear(config.hidden_size, self.all_head_size)
        self.key = nn.Linear(config.hidden_size, self.all_head_size)
        self.value = nn.Linear(config.hidden_size, self.all_head_size)

        self.dropout = nn.Dropout(config.attention_probs_dropout_prob)
        self.softmax = nn.Softmax(dim=-1)

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, self.attention_head_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def transpose_key_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, self.attention_head_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 3, 1)

    def forward(self, hidden_states, attention_mask):
        mixed_query_layer = self.query(hidden_states)
        mixed_key_layer = self.key(hidden_states)
        mixed_value_layer = self.value(hidden_states)

        query_layer = self.transpose_for_scores(mixed_query_layer)
        key_layer = self.transpose_key_for_scores(mixed_key_layer)
        value_layer = self.transpose_for_scores(mixed_value_layer)

        # Take the dot product between "query" and "key" to get the raw attention scores.
        attention_scores = torch.matmul(query_layer, key_layer)
        attention_scores = attention_scores / math.sqrt(self.attention_head_size)
        # Apply the attention mask is (precomputed for all layers in BertModel forward() function)
        attention_scores = attention_scores + attention_mask

        # Normalize the attention scores to probabilities.
        attention_probs = self.softmax(attention_scores)

        # This is actually dropping out entire tokens to attend to, which might
        # seem a bit unusual, but is taken from the original Transformer paper.
        attention_probs = self.dropout(attention_probs)

        context_layer = torch.matmul(attention_probs, value_layer)
        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(*new_context_layer_shape)
        return context_layer


class BertSelfOutput(nn.Module):
    def __init__(self, config):
        super(BertSelfOutput, self).__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.dense.bert_output_layer = True
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, hidden_states, input_tensor):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        return hidden_states


class BertAttention(nn.Module):
    def __init__(self, config):
        super(BertAttention, self).__init__()
        self.self = BertSelfAttention(config)
        self.output = BertSelfOutput(config)

    def forward(self, input_tensor, attention_mask):
        self_output = self.self(input_tensor, attention_mask)
        attention_output = self.output(self_output, input_tensor)
        return attention_output


class BertIntermediate(nn.Module):
    def __init__(self, config):
        super(BertIntermediate, self).__init__()
        self.dense_act = LinearActivation(config.hidden_size, config.intermediate_size, act=config.hidden_act)

    def forward(self, hidden_states):
        hidden_states = self.dense_act(hidden_states)
        return hidden_states


class BertOutput(nn.Module):
    def __init__(self, config):
        super(BertOutput, self).__init__()
        self.dense = nn.Linear(config.intermediate_size, config.hidden_size)
        self.dense.bert_output_layer = True
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, hidden_states):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        return hidden_states


class BertLayer(nn.Module):
    def __init__(self, config):
        super(BertLayer, self).__init__()
        self.attention = BertAttention(config)
        self.PreAttentionLayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)
        self.PostAttentionLayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)
        self.intermediate = BertIntermediate(config)
        self.output = BertOutput(config)

    def forward(self, hidden_states, attention_mask):
        input_layer_norm = self.PreAttentionLayerNorm(hidden_states)
        attention_output = self.attention(input_layer_norm, attention_mask)

        intermediate_input = hidden_states + attention_output

        intermediate_layer_norm = self.PostAttentionLayerNorm(intermediate_input)
        intermediate_output = self.intermediate(intermediate_layer_norm)
        layer_output = self.output(intermediate_output)

        return layer_output + intermediate_input


class BertEncoder(nn.Module):
    def __init__(self, config):
        super(BertEncoder, self).__init__()

        # Added later to make it similar to GPT-2
        self.FinalLayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)

        layer = BertLayer(config)
        self.layer = nn.ModuleList([copy.deepcopy(layer) for _ in range(config.num_hidden_layers)])

    def forward(self, hidden_states, attention_mask, output_all_encoded_layers=True, checkpoint_activations=False):
        all_encoder_layers = []

        def custom(start, end):
            def custom_forward(*inputs):
                layers = self.layer[start:end]
                x_ = inputs[0]
                for layer in layers:
                    x_ = layer(x_, inputs[1])
                return x_

            return custom_forward

        if checkpoint_activations:
            l = 0  # noqa E741
            num_layers = len(self.layer)
            chunk_length = math.ceil(math.sqrt(num_layers))
            while l < num_layers:
                hidden_states = checkpoint.checkpoint(custom(l, l + chunk_length), hidden_states, attention_mask * 1)
                l += chunk_length  # noqa E741
            # decoder layers
        else:
            for i, layer_module in enumerate(self.layer):
                hidden_states = layer_module(hidden_states, attention_mask)

                if output_all_encoded_layers:
                    all_encoder_layers.append(hidden_states)

        if not output_all_encoded_layers or checkpoint_activations:
            hidden_states = self.FinalLayerNorm(hidden_states)
            all_encoder_layers.append(hidden_states)
        return all_encoder_layers


class BertPooler(nn.Module):
    def __init__(self, config):
        super(BertPooler, self).__init__()
        self.dense_act = LinearActivation(config.hidden_size, config.hidden_size, act="tanh")

    def forward(self, hidden_states):
        # We "pool" the model by simply taking the hidden state corresponding
        # to the first token.
        first_token_tensor = hidden_states[:, 0]
        pooled_output = self.dense_act(first_token_tensor)
        return pooled_output


class BertPredictionHeadTransform(nn.Module):
    def __init__(self, config):
        super(BertPredictionHeadTransform, self).__init__()
        self.dense_act = LinearActivation(config.hidden_size, config.hidden_size, act=config.hidden_act)
        self.LayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)

    def forward(self, hidden_states):
        hidden_states = self.dense_act(hidden_states)
        hidden_states = self.LayerNorm(hidden_states)
        return hidden_states


class BertLMPredictionHead(nn.Module):
    def __init__(self, config, bert_model_embedding_weights):
        super(BertLMPredictionHead, self).__init__()
        self.transform = BertPredictionHeadTransform(config)

        # The output weights are the same as the input embeddings, but there is
        # an output-only bias for each token.
        self.decoder = nn.Linear(bert_model_embedding_weights.size(1), bert_model_embedding_weights.size(0), bias=False)
        self.decoder.weight = bert_model_embedding_weights
        self.bias = nn.Parameter(torch.zeros(bert_model_embedding_weights.size(0)))

    def forward(self, hidden_states, masked_token_indexes):
        hidden_states = self.transform(hidden_states)

        if masked_token_indexes is not None:
            hidden_states = torch.index_select(hidden_states.view(-1, hidden_states.shape[-1]), 0, masked_token_indexes)

        hidden_states = self.decoder(hidden_states) + self.bias
        return hidden_states


class BertPreTrainingHeads(nn.Module):
    def __init__(self, config, bert_model_embedding_weights):
        super(BertPreTrainingHeads, self).__init__()
        self.predictions = BertLMPredictionHead(config, bert_model_embedding_weights)
        self.seq_relationship = nn.Linear(config.hidden_size, 2)

    def forward(self, sequence_output, pooled_output, masked_token_indexes=None):
        prediction_scores = self.predictions(sequence_output, masked_token_indexes)
        seq_relationship_score = self.seq_relationship(pooled_output)
        return prediction_scores, seq_relationship_score


class BertPreTrainedModel(nn.Module):
    """An abstract class to handle weights initialization and
    a simple interface for dowloading and loading pretrained models.
    """

    def __init__(self, config, *inputs, **kwargs):
        super(BertPreTrainedModel, self).__init__()
        if not isinstance(config, BertConfig):
            raise ValueError(
                "Parameter config in `{}(config)` should be an instance of class `BertConfig`. "
                "To create a model from a Google pretrained model use "
                "`model = {}.from_pretrained(PRETRAINED_MODEL_NAME)`".format(
                    self.__class__.__name__, self.__class__.__name__
                )
            )
        self.config = config

    def init_bert_weights(self, module):
        """Initialize the weights."""
        if isinstance(module, (nn.Linear, nn.Embedding)):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            num_layers = self.config.num_hidden_layers
            std = self.config.initializer_range
            if hasattr(module, "bert_output_layer"):
                # "Accounting for accumulation on the residual path"
                # print("Accounting for accumulation on the residual path")
                std = self.config.initializer_range / math.sqrt(2.0 * num_layers)
            module.weight.data.normal_(mean=0.0, std=std)
        elif isinstance(module, BertLayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        if isinstance(module, nn.Linear) and module.bias is not None:
            module.bias.data.zero_()


class BertModel(BertPreTrainedModel):
    """BERT model ("Bidirectional Embedding Representations from a Transformer").

    Params:
        config: a BertConfig class instance with the configuration to build a new model

    Inputs:
        `input_ids`: a torch.LongTensor of shape [batch_size, sequence_length]
            with the word token indices in the vocabulary(see the tokens preprocessing logic in the scripts
            `extract_features.py`, `run_classifier.py` and `run_squad.py`)
        `token_type_ids`: an optional torch.LongTensor of shape [batch_size, sequence_length] with the token
            types indices selected in [0, 1]. Type 0 corresponds to a `sentence A` and type 1 corresponds to
            a `sentence B` token (see BERT paper for more details).
        `attention_mask`: an optional torch.LongTensor of shape [batch_size, sequence_length] with indices
            selected in [0, 1]. It's a mask to be used if the input sequence length is smaller than the max
            input sequence length in the current batch. It's the mask that we typically use for attention when
            a batch has varying length sentences.
        `output_all_encoded_layers`: boolean which controls the content of the `encoded_layers` output as described below. Default: `True`.

    Outputs: Tuple of (encoded_layers, pooled_output)
        `encoded_layers`: controled by `output_all_encoded_layers` argument:
            - `output_all_encoded_layers=True`: outputs a list of the full sequences of encoded-hidden-states at the end
                of each attention block (i.e. 12 full sequences for BERT-base, 24 for BERT-large), each
                encoded-hidden-state is a torch.FloatTensor of size [batch_size, sequence_length, hidden_size],
            - `output_all_encoded_layers=False`: outputs only the full sequence of hidden-states corresponding
                to the last attention block of shape [batch_size, sequence_length, hidden_size],
        `pooled_output`: a torch.FloatTensor of size [batch_size, hidden_size] which is the output of a
            classifier pretrained on top of the hidden state associated to the first character of the
            input (`CLS`) to train on the Next-Sentence task (see BERT's paper).

    Example usage:
    ```python
    # Already been converted into WordPiece token ids
    input_ids = torch.LongTensor([[31, 51, 99], [15, 5, 0]])
    input_mask = torch.LongTensor([[1, 1, 1], [1, 1, 0]])
    token_type_ids = torch.LongTensor([[0, 0, 1], [0, 1, 0]])

    config = modeling.BertConfig(vocab_size_or_config_json_file=32000, hidden_size=768,
        num_hidden_layers=12, num_attention_heads=12, intermediate_size=3072)

    model = modeling.BertModel(config=config)
    all_encoder_layers, pooled_output = model(input_ids, token_type_ids, input_mask)
    ```
    """

    def __init__(self, config):
        super(BertModel, self).__init__(config)
        self.embeddings = BertEmbeddings(config)
        self.encoder = BertEncoder(config)
        self.pooler = BertPooler(config)
        self.apply(self.init_bert_weights)
        logger.info("Init BERT pretrain model")

    def forward(
        self,
        input_ids,
        token_type_ids=None,
        attention_mask=None,
        output_all_encoded_layers=True,
        checkpoint_activations=False,
    ):
        if attention_mask is None:
            attention_mask = torch.ones_like(input_ids)
        if token_type_ids is None:
            token_type_ids = torch.zeros_like(input_ids)

        # We create a 3D attention mask from a 2D tensor mask.
        # Sizes are [batch_size, 1, 1, to_seq_length]
        # So we can broadcast to [batch_size, num_heads, from_seq_length, to_seq_length]
        # this attention mask is more simple than the triangular masking of causal attention
        # used in OpenAI GPT, we just need to prepare the broadcast dimension here.
        extended_attention_mask = attention_mask.unsqueeze(1).unsqueeze(2)

        # Since attention_mask is 1.0 for positions we want to attend and 0.0 for
        # masked positions, this operation will create a tensor which is 0.0 for
        # positions we want to attend and -10000.0 for masked positions.
        # Since we are adding it to the raw scores before the softmax, this is
        # effectively the same as removing these entirely.
        extended_attention_mask = extended_attention_mask.to(dtype=next(self.parameters()).dtype)  # fp16 compatibility
        extended_attention_mask = (1.0 - extended_attention_mask) * -10000.0

        embedding_output = self.embeddings(input_ids, token_type_ids)
        encoded_layers = self.encoder(
            embedding_output,
            extended_attention_mask,
            output_all_encoded_layers=output_all_encoded_layers,
            checkpoint_activations=checkpoint_activations,
        )
        sequence_output = encoded_layers[-1]
        pooled_output = self.pooler(sequence_output)

        if not output_all_encoded_layers:
            encoded_layers = encoded_layers[-1]
        return encoded_layers, pooled_output


class BertForPreTrainingPreLN(BertPreTrainedModel):
    """BERT model with pre-training heads.
    Params:
        config: a BertConfig class instance with the configuration to build a new model.

    """

    def __init__(self, config):
        super(BertForPreTrainingPreLN, self).__init__(config)
        self.bert = BertModel(config)
        self.cls = BertPreTrainingHeads(config, self.bert.embeddings.word_embeddings.weight)
        self.apply(self.init_bert_weights)



import math

import torch
import torch.nn as nn

from cogdl.utils import spmm, get_activation


# def propagation(graph, x, type="gcn"):
#     hidden = spmm(graph, x)

#     row, col = graph.edge_index
#     # Self-attention on the nodes - Shared attention mechanism
#     h_l = (self.a_l * h).sum(dim=-1)
#     h_r = (self.a_r * h).sum(dim=-1)

#     if self.dropout.p == 0.0 and graph.is_symmetric() and check_fused_gat():
#         out = fused_gat_op(h_l, h_r, graph, self.alpha, h)
#         out = out.view(out.shape[0], -1)
#     else:
#         # edge_attention: E * H
#         edge_attention = self.leakyrelu(h_l[row] + h_r[col])
#         edge_attention = edge_softmax(graph, edge_attention)
#         edge_attention = self.dropout(edge_attention)

#         if check_mh_spmm() and next(self.parameters()).device.type != "cpu":
#             if self.nhead > 1:
#                 h_prime = mh_spmm(graph, edge_attention, h)
#                 out = h_prime.view(h_prime.shape[0], -1)
#             else:
#                 edge_weight = edge_attention.view(-1)
#                 with graph.local_graph():
#                     graph.edge_weight = edge_weight
#                     out = spmm(graph, h.squeeze(1))
#         else:
#             with graph.local_graph():
#                 h_prime = []
#                 h = h.permute(1, 0, 2).contiguous()
#                 for i in range(self.nhead):
#                     edge_weight = edge_attention[:, i]
#                     graph.edge_weight = edge_weight
#                     hidden = h[i]
#                     assert not torch.isnan(hidden).any()
#                     h_prime.append(spmm(graph, hidden))
#             out = torch.cat(h_prime, dim=1)


class AutoLayer(nn.Module):

    def __init__(
        self, in_features, out_features, dropout=0.0, activation=None, residual=False, norm=None, bias=True, alpha=0.1, beta=1, **kwargs
    ):
        super(AutoLayer, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.alpha = alpha
        self.beta = beta
        self.linear = nn.Linear(in_features, out_features, bias=bias)
        if dropout > 0:
            self.dropout = nn.Dropout(dropout)
        else:
            self.dropout = None
        if residual:
            self.residual = nn.Linear(in_features, out_features)
        else:
            self.residual = None

        if activation is not None:
            self.act = get_activation(activation, inplace=True)
        else:
            self.act = None

        if norm is not None:
            if norm == "batchnorm":
                self.norm = nn.BatchNorm1d(out_features)
            elif norm == "layernorm":
                self.norm = nn.LayerNorm(out_features)
            else:
                raise NotImplementedError
        else:
            self.norm = None

        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.out_features)
        torch.nn.init.uniform_(self.linear.weight, -stdv, stdv)

    def forward(self, graph, x, init_x):
        hidden = spmm(graph, x)
        hidden = (1 - self.alpha) * hidden + self.alpha * init_x
        out = self.beta * self.linear(hidden) + (1 - self.beta) * hidden

        if self.norm is not None:
            out = self.norm(out)
        if self.act is not None:
            out = self.act(out)

        if self.residual is not None:
            out = out + self.residual(x)
        if self.dropout is not None:
            out = self.dropout(out)
        return out



import torch
from torch import nn
from torch.nn import CrossEntropyLoss
import logging
from .bert_model import BertPreTrainedModel, BertPreTrainingHeads, BertModel, BertEncoder, BertPooler, BertLayerNorm


logger = logging.getLogger(__name__)


class DualPositionBertEmbeddings(nn.Module):
    """Construct the embeddings from word, position and token_type embeddings."""

    def __init__(self, config):
        super(DualPositionBertEmbeddings, self).__init__()
        self.word_embeddings = nn.Embedding(config.vocab_size, config.hidden_size)
        self.position_embeddings = nn.Embedding(config.max_position_embeddings, config.hidden_size)
        self.position_embeddings_second = nn.Embedding(config.max_position_embeddings, config.hidden_size)
        self.token_type_embeddings = nn.Embedding(config.type_vocab_size, config.hidden_size)
        # self.LayerNorm is not snake-cased to stick with TensorFlow model variable name and be able to load
        # any TensorFlow checkpoint file
        self.LayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, input_ids, token_type_ids, position_ids, position_ids_second):
        if token_type_ids is None:
            token_type_ids = torch.zeros_like(input_ids)

        words_embeddings = self.word_embeddings(input_ids)
        position_embeddings = self.position_embeddings(position_ids)
        position_embeddings_second = self.position_embeddings(position_ids_second)
        token_type_embeddings = self.token_type_embeddings(token_type_ids)

        embeddings = words_embeddings + position_embeddings + position_embeddings_second + token_type_embeddings
        embeddings = self.LayerNorm(embeddings)
        embeddings = self.dropout(embeddings)
        return embeddings


class DualPositionBertModel(BertModel):
    def __init__(self, config):
        super(DualPositionBertModel, self).__init__(config)
        self.embeddings = DualPositionBertEmbeddings(config)
        self.encoder = BertEncoder(config)
        self.pooler = BertPooler(config)
        self.apply(self.init_bert_weights)
        logger.info("Init BERT pretrain model")

    def forward(
        self,
        input_ids,
        token_type_ids=None,
        attention_mask=None,
        output_all_encoded_layers=True,
        checkpoint_activations=False,
        position_ids=None,
        position_ids_second=None,
    ):
        if attention_mask is None:
            attention_mask = torch.ones_like(input_ids)
        if token_type_ids is None:
            token_type_ids = torch.zeros_like(input_ids)

        if len(attention_mask.shape) == 2:
            extended_attention_mask = attention_mask.unsqueeze(1).unsqueeze(2)
        elif len(attention_mask.shape) == 3:
            extended_attention_mask = attention_mask.unsqueeze(1)
        else:
            raise Exception("invalid attention mask shape! shape: %s" % (attention_mask.shape))

        extended_attention_mask = extended_attention_mask.to(dtype=next(self.parameters()).dtype)  # fp16 compatibility
        extended_attention_mask = (1.0 - extended_attention_mask) * -10000.0
        embedding_output = self.embeddings(input_ids, token_type_ids, position_ids, position_ids_second)
        encoded_layers = self.encoder(
            embedding_output,
            extended_attention_mask,
            output_all_encoded_layers=output_all_encoded_layers,
            checkpoint_activations=checkpoint_activations,
        )
        sequence_output = encoded_layers[-1]
        pooled_output = self.pooler(sequence_output)

        if not output_all_encoded_layers:
            encoded_layers = encoded_layers[-1]
        return encoded_layers, pooled_output


class DualPositionBertForPreTrainingPreLN(BertPreTrainedModel):
    """BERT model with pre-training heads and dual position
    Params:
        config: a BertConfig class instance with the configuration to build a new model.

    """

    def __init__(self, config):
        super(DualPositionBertForPreTrainingPreLN, self).__init__(config)
        self.bert = DualPositionBertModel(config)
        self.cls = BertPreTrainingHeads(config, self.bert.embeddings.word_embeddings.weight)
        self.apply(self.init_bert_weights)

    def forward(
        self,
        input_ids,
        token_type_ids=None,
        attention_mask=None,
        masked_lm_labels=None,
        position_ids=None,
        position_ids_second=None,
        log=True,
    ):
        sequence_output, pooled_output = self.bert(
            input_ids=input_ids,
            token_type_ids=token_type_ids,
            attention_mask=attention_mask,
            output_all_encoded_layers=False,
            checkpoint_activations=False,
            position_ids=position_ids,
            position_ids_second=position_ids_second,
        )

        if masked_lm_labels is not None:
            # filter out all masked labels.
            masked_token_indexes = torch.nonzero((masked_lm_labels + 1).view(-1)).view(-1)
            prediction_scores, _ = self.cls(sequence_output, pooled_output, masked_token_indexes)
            target = torch.index_select(masked_lm_labels.view(-1), 0, masked_token_indexes)

            loss_fct = CrossEntropyLoss(ignore_index=-1)
            masked_lm_loss = loss_fct(prediction_scores.view(-1, self.config.vocab_size), target)
            return masked_lm_loss
        else:
            prediction_scores, _ = self.cls(sequence_output, pooled_output)
            return prediction_scores



import torch
import torch.nn as nn
import torch.nn.functional as F
from cogdl.utils.srgcn_utils import act_attention, act_map, act_normalization
from cogdl.utils import spmm, add_remaining_self_loops
from torch_sparse import spspmm

from .. import BaseModel


class NodeAdaptiveEncoder(nn.Module):
    def __init__(self, num_features, dropout=0.5):
        super(NodeAdaptiveEncoder, self).__init__()
        self.fc = nn.Parameter(torch.zeros(size=(num_features, 1)))
        nn.init.xavier_normal_(self.fc.data, gain=1.414)
        self.bf = nn.Parameter(torch.zeros(size=(1,)))

        self.dropout = torch.nn.Dropout(dropout)

    def forward(self, x):
        h = torch.mm(x, self.fc) + self.bf
        h = torch.sigmoid(h)
        h = self.dropout(h)

        return torch.where(x < 0, torch.zeros_like(x), x) + h * torch.where(x > 0, torch.zeros_like(x), x)


class SrgcnHead(nn.Module):
    def __init__(
        self,
        num_features,
        out_feats,
        attention,
        activation,
        normalization,
        nhop,
        subheads=2,
        dropout=0.5,
        node_dropout=0.5,
        alpha=0.2,
        concat=True,
    ):
        super(SrgcnHead, self).__init__()

        self.subheads = subheads
        self.concat = concat
        self.alpha = alpha
        self.nhop = nhop

        self.attention = attention(out_feats)
        self.activation = activation
        self.normalization = normalization()

        self.adaptive_enc = nn.ModuleList()

        self.weight = nn.ParameterList()
        self.bias = nn.ParameterList()

        # multi-head
        for r in range(self.subheads):
            W = nn.Parameter(torch.zeros(size=(num_features, out_feats)))
            nn.init.xavier_normal_(W.data, gain=1.414)
            self.weight.append(W)
            self.bias.append(nn.Parameter(torch.zeros(size=(out_feats,))))
            self.adaptive_enc.append(NodeAdaptiveEncoder(out_feats, dropout))

        self.dropout = dropout
        self.node_dropout = node_dropout

    def forward(self, graph, x):
        edge_index = graph.edge_index
        N, dim = x.shape

        nl_adj_mat_ind = add_remaining_self_loops(edge_index, num_nodes=N)[0]
        nl_adj_mat_ind = torch.stack(nl_adj_mat_ind)
        nl_adj_mat_val = torch.ones(nl_adj_mat_ind.shape[1]).to(x.device)

        for _ in range(self.nhop - 1):
            nl_adj_mat_ind, nl_adj_mat_val = spspmm(
                nl_adj_mat_ind, nl_adj_mat_val, nl_adj_mat_ind, nl_adj_mat_val, N, N, N, True
            )

        result = []
        for i in range(self.subheads):
            h = torch.mm(x, self.weight[i])

            adj_mat_ind, adj_mat_val = nl_adj_mat_ind, nl_adj_mat_val
            h = F.dropout(h, p=self.dropout, training=self.training)

            adj_mat_ind, adj_mat_val = self.attention(h, adj_mat_ind, adj_mat_val)
            # laplacian matrix normalization
            adj_mat_val = self.normalization(adj_mat_ind, adj_mat_val, N)

            val_h = h

            with graph.local_graph():
                graph.edge_index = adj_mat_ind
                graph.edge_weight = adj_mat_val
                for _ in range(i + 1):
                    val_h = spmm(graph, val_h)

                val_h[val_h != val_h] = 0
                val_h = val_h + self.bias[i]
                val_h = self.adaptive_enc[i](val_h)
                val_h = self.activation(val_h)
                val_h = F.dropout(val_h, p=self.dropout, training=self.training)
                result.append(val_h)
        h_res = torch.cat(result, dim=1)
        return h_res


class SrgcnSoftmaxHead(nn.Module):
    def __init__(
        self,
        num_features,
        out_feats,
        attention,
        activation,
        nhop,
        normalization,
        dropout=0.5,
        node_dropout=0.5,
        alpha=0.2,
    ):
        super(SrgcnSoftmaxHead, self).__init__()

        self.alpha = alpha
        self.activation = activation
        self.nhop = nhop
        self.normalization = normalization()
        self.attention = attention(out_feats)

        self.weight = nn.Parameter(torch.zeros(size=(num_features, out_feats)))
        nn.init.xavier_normal_(self.weight.data, gain=1.414)
        self.bias = nn.Parameter(torch.zeros(size=(out_feats,)))
        self.adaptive_enc = NodeAdaptiveEncoder(out_feats, dropout)

        self.dropout = dropout
        self.node_dropout = node_dropout

    def forward(self, graph, x):
        N, dim = x.shape
        # x = self.dropout(x)

        edge_index = graph.edge_index
        adj_mat_ind = add_remaining_self_loops(edge_index, num_nodes=N)[0]
        adj_mat_ind = torch.stack(adj_mat_ind)
        adj_mat_val = torch.ones(adj_mat_ind.shape[1]).to(x.device)

        h = torch.mm(x, self.weight)
        h = F.dropout(h, p=self.dropout, training=self.training)
        for _ in range(self.nhop - 1):
            adj_mat_ind, adj_mat_val = spspmm(adj_mat_ind, adj_mat_val, adj_mat_ind, adj_mat_val, N, N, N, True)

        adj_mat_ind, adj_mat_val = self.attention(h, adj_mat_ind, adj_mat_val)

        # MATRIX_MUL
        # laplacian matrix normalization
        adj_mat_val = self.normalization(adj_mat_ind, adj_mat_val, N)

        val_h = h
        # N, dim = val_h.shape

        # MATRIX_MUL
        with graph.local_graph():
            graph.edge_index = adj_mat_ind
            graph.edge_weight = adj_mat_val
            val_h = spmm(graph, val_h)

        val_h[val_h != val_h] = 0
        val_h = val_h + self.bias
        val_h = self.adaptive_enc(val_h)
        val_h = F.dropout(val_h, p=self.dropout, training=self.training)
        # val_h = self.activation(val_h)
        return val_h


class SRGCN(BaseModel):
    @staticmethod
    def add_args(parser):
        parser.add_argument("--hidden-size", type=int, default=8)
        parser.add_argument("--num-heads", type=int, default=8)
        parser.add_argument("--dropout", type=float, default=0.5)
        parser.add_argument("--node-dropout", type=float, default=0.5)
        parser.add_argument("--alpha", type=float, default=0.2)
        parser.add_argument("--lr", type=float, default=0.005)
        parser.add_argument("--subheads", type=int, default=1)
        parser.add_argument("--attention-type", type=str, default="node")
        parser.add_argument("--activation", type=str, default="leaky_relu")
        parser.add_argument("--nhop", type=int, default=1)
        parser.add_argument("--normalization", type=str, default="row_uniform")

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            in_feats=args.num_features,
            hidden_size=args.hidden_size,
            out_feats=args.num_classes,
            dropout=args.dropout,
            node_dropout=args.node_dropout,
            nhead=args.num_heads,
            subheads=args.subheads,
            alpha=args.alpha,
            attention=args.attention_type,
            activation=args.activation,
            nhop=args.nhop,
            normalization=args.normalization,
        )

    def __init__(
        self,
        in_feats,
        hidden_size,
        out_feats,
        attention,
        activation,
        nhop,
        normalization,
        dropout,
        node_dropout,
        alpha,
        nhead,
        subheads,
    ):
        super(SRGCN, self).__init__()
        attn_f = act_attention(attention)
        activate_f = act_map(activation)
        norm_f = act_normalization(normalization)
        self.attentions = [
            SrgcnHead(
                num_features=in_feats,
                out_feats=hidden_size,
                attention=attn_f,
                activation=activate_f,
                nhop=nhop,
                normalization=norm_f,
                subheads=subheads,
                dropout=dropout,
                node_dropout=node_dropout,
                alpha=alpha,
                concat=True,
            )
            for _ in range(nhead)
        ]
        for i, attention in enumerate(self.attentions):
            self.add_module("attention_{}".format(i), attention)
        self.out_att = SrgcnSoftmaxHead(
            num_features=hidden_size * nhead * subheads,
            out_feats=out_feats,
            attention=attn_f,
            activation=activate_f,
            normalization=act_normalization("row_softmax"),
            nhop=nhop,
            dropout=dropout,
            node_dropout=node_dropout,
        )

    def forward(self, graph):
        x = torch.cat([att(graph, graph.x) for att in self.attentions], dim=1)
        x = F.elu(x)
        x = self.out_att(graph, x)
        return x

    def predict(self, data):
        return self.forward(data)



from abc import ABCMeta
import torch.utils.data
from torch.utils.data.dataloader import default_collate

from cogdl.data import Batch, Graph

try:
    from typing import GenericMeta  # python 3.6
except ImportError:
    # in 3.7, genericmeta doesn't exist but we don't need it
    class GenericMeta(type):
        pass


class RecordParameters(ABCMeta):
    def __call__(cls, *args, **kwargs):
        obj = type.__call__(cls, *args, **kwargs)
        obj.record_parameters([args, kwargs])
        return obj


class GenericRecordParameters(GenericMeta, RecordParameters):
    pass


class DataLoader(torch.utils.data.DataLoader, metaclass=GenericRecordParameters):
    r"""Data loader which merges data objects from a
    :class:`cogdl.data.dataset` to a mini-batch.

    Args:
        dataset (Dataset): The dataset from which to load the data.
        batch_size (int, optional): How may samples per batch to load.
            (default: :obj:`1`)
        shuffle (bool, optional): If set to :obj:`True`, the data will be
            reshuffled at every epoch (default: :obj:`True`)
    """

    def __init__(self, dataset, batch_size=1, shuffle=True, **kwargs):
        if "collate_fn" not in kwargs or kwargs["collate_fn"] is None:
            kwargs["collate_fn"] = self.collate_fn

        super(DataLoader, self).__init__(
            dataset,
            batch_size,
            shuffle,
            **kwargs,
        )

    @staticmethod
    def collate_fn(batch):
        item = batch[0]
        if isinstance(item, Graph):
            return Batch.from_data_list(batch)
        elif isinstance(item, torch.Tensor):
            return default_collate(batch)
        elif isinstance(item, float):
            return torch.tensor(batch, dtype=torch.float)

        raise TypeError("DataLoader found invalid type: {}".format(type(item)))

    def get_parameters(self):
        return self.default_kwargs

    def record_parameters(self, params):
        self.default_kwargs = params



import math

import torch
import torch.nn as nn

from actnn.layers import QLinear

from cogdl.utils import spmm


class ActGCNIILayer(nn.Module):
    def __init__(self, n_channels, alpha=0.1, beta=1, residual=False):
        super(ActGCNIILayer, self).__init__()
        self.n_channels = n_channels
        self.alpha = alpha
        self.beta = beta
        self.residual = residual
        self.linear = QLinear(n_channels, n_channels)
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.n_channels)
        torch.nn.init.uniform_(self.linear.weight, -stdv, stdv)

    def forward(self, graph, x, init_x):
        """Symmetric normalization"""
        hidden = spmm(graph, x, actnn=True)
        hidden = (1 - self.alpha) * hidden + self.alpha * init_x
        h = self.beta * self.linear(hidden) + (1 - self.beta) * hidden
        if self.residual:
            h = h + x
        return h



import torch.nn as nn

from actnn.layers import QLinear, QReLU, QBatchNorm1d, QDropout


class ActMLP(nn.Module):
    r"""Multilayer perception with normalization

    .. math::
        x^{(i+1)} = \sigma(W^{i}x^{(i)})

    Parameters
    ----------
    in_feats : int
        Size of each input sample.
    out_feats : int
        Size of each output sample.
    hidden_dim : int
        Size of hidden layer dimension.
    use_bn : bool, optional
        Apply batch normalization if True, default: `True).
    """

    def __init__(
        self,
        in_feats,
        out_feats,
        hidden_size,
        num_layers,
        dropout=0.0,
        activation="relu",
        norm=None,
        act_first=False,
        bias=True,
    ):
        super(ActMLP, self).__init__()
        self.norm = norm
        if activation == "relu":
            self.activation = QReLU()
        else:
            self.activation = lambda x: x
        self.act_first = act_first
        if dropout > 0.0:
            self.dropout = QDropout(dropout)
        else:
            self.dropout = None
        shapes = [in_feats] + [hidden_size] * (num_layers - 1) + [out_feats]
        self.mlp = nn.ModuleList([QLinear(shapes[layer], shapes[layer + 1], bias=bias) for layer in range(num_layers)])
        if norm is not None and num_layers > 1:
            if norm == "batchnorm":
                self.norm_list = nn.ModuleList(QBatchNorm1d(x) for x in shapes[1:-1])
            else:
                raise NotImplementedError(f"{norm} is not implemented in CogDL.")
        self.reset_parameters()

    def reset_parameters(self):
        for layer in self.mlp:
            layer.reset_parameters()
        if hasattr(self, "norm_list"):
            for n in self.norm_list:
                n.reset_parameters()

    def forward(self, x):
        for i, fc in enumerate(self.mlp[:-1]):
            x = fc(x)
            if self.act_first:
                x = self.activation(x)
            if self.norm:
                x = self.norm_list[i](x)

            if not self.act_first:
                x = self.activation(x)
            if self.dropout is not None:
                x = self.dropout(x)
        x = self.mlp[-1](x)
        return x



import os.path as osp

import torch
from cogdl.data import Graph, Dataset
from cogdl.utils import download_url


def read_triplet_data(folder):
    filenames = ["train2id.txt", "valid2id.txt", "test2id.txt"]
    count = 0
    edge_index = []
    edge_attr = []
    count_list = []
    triples = []
    num_entities = 0
    num_relations = 0
    entity_dic = {}
    relation_dic = {}
    for filename in filenames:
        with open(osp.join(folder, filename), "r") as f:
            _ = int(f.readline().strip())
            if "train" in filename:
                train_start_idx = len(triples)
            elif "valid" in filename:
                valid_start_idx = len(triples)
            elif "test" in filename:
                test_start_idx = len(triples)
            for line in f:
                items = line.strip().split()
                edge_index.append([int(items[0]), int(items[1])])
                edge_attr.append(int(items[2]))
                triples.append((int(items[0]), int(items[2]), int(items[1])))
                if items[0] not in entity_dic:
                    entity_dic[items[0]] = num_entities
                    num_entities += 1
                if items[1] not in entity_dic:
                    entity_dic[items[1]] = num_entities
                    num_entities += 1
                if items[2] not in relation_dic:
                    relation_dic[items[2]] = num_relations
                    num_relations += 1
                count += 1
            count_list.append(count)

    edge_index = torch.LongTensor(edge_index).t()
    edge_attr = torch.LongTensor(edge_attr)
    data = Graph()
    data.edge_index = edge_index
    data.edge_attr = edge_attr

    def generate_mask(start, end):
        mask = torch.BoolTensor(count)
        mask[:] = False
        mask[start:end] = True
        return mask

    data.train_mask = generate_mask(0, count_list[0])
    data.val_mask = generate_mask(count_list[0], count_list[1])
    data.test_mask = generate_mask(count_list[1], count_list[2])
    return data, triples, train_start_idx, valid_start_idx, test_start_idx, num_entities, num_relations


class KnowledgeGraphDataset(Dataset):
    # url = "https://raw.githubusercontent.com/thunlp/OpenKE/OpenKE-PyTorch/benchmarks"
    url = "https://cloud.tsinghua.edu.cn/d/b567292338f2488699b7/files/?p=%2F{}%2F{}&dl=1"

    def __init__(self, root, name):
        self.name = name
        super(KnowledgeGraphDataset, self).__init__(root)
        self.data = torch.load(self.processed_paths[0])
        triple_config = torch.load(self.processed_paths[1])
        self.triples = triple_config["triples"]
        self._train_start_index = triple_config["train_start_index"]
        self._valid_start_index = triple_config["valid_start_index"]
        self._test_start_index = triple_config["test_start_index"]
        self._num_entities = triple_config["num_entities"]
        self._num_relations = triple_config["num_relations"]

    @property
    def raw_file_names(self):
        names = ["train2id.txt", "valid2id.txt", "test2id.txt"]
        return names

    @property
    def processed_file_names(self):
        return ["data.pt", "triple_config.pt"]

    @property
    def train_start_idx(self):
        return self._train_start_index

    @property
    def valid_start_idx(self):
        return self._valid_start_index

    @property
    def test_start_idx(self):
        return self._test_start_index

    @property
    def num_entities(self):
        return self._num_entities

    @property
    def num_relations(self):
        return self._num_relations

    def get(self, idx):
        assert idx == 0
        return self.data

    def download(self):
        for name in self.raw_file_names:
            # download_url("{}/{}/{}".format(self.url, self.name, name), self.raw_dir)
            download_url(self.url.format(self.name, name), self.raw_dir, name=name)

    def process(self):
        (
            data,
            triples,
            train_start_index,
            valid_start_index,
            test_start_index,
            num_entities,
            num_relations,
        ) = read_triplet_data(self.raw_dir)
        torch.save(data, self.processed_paths[0])
        triple_config = {
            "triples": triples,
            "train_start_index": train_start_index,
            "valid_start_index": valid_start_index,
            "test_start_index": test_start_index,
            "num_entities": num_entities,
            "num_relations": num_relations,
        }
        torch.save(triple_config, self.processed_paths[1])


class FB13Datset(KnowledgeGraphDataset):
    def __init__(self, data_path="data"):
        dataset = "FB13"
        path = osp.join(data_path, dataset)
        super(FB13Datset, self).__init__(path, dataset)


class FB15kDatset(KnowledgeGraphDataset):
    def __init__(self, data_path="data"):
        dataset = "FB15K"
        path = osp.join(data_path, dataset)
        super(FB15kDatset, self).__init__(path, dataset)


class FB15k237Datset(KnowledgeGraphDataset):
    def __init__(self, data_path="data"):
        dataset = "FB15K237"
        path = osp.join(data_path, dataset)
        super(FB15k237Datset, self).__init__(path, dataset)


class WN18Datset(KnowledgeGraphDataset):
    def __init__(self, data_path="data"):
        dataset = "WN18"
        path = osp.join(data_path, dataset)
        super(WN18Datset, self).__init__(path, dataset)


class WN18RRDataset(KnowledgeGraphDataset):
    def __init__(self, data_path="data"):
        dataset = "WN18RR"
        path = osp.join(data_path, dataset)
        super(WN18RRDataset, self).__init__(path, dataset)


class FB13SDatset(KnowledgeGraphDataset):
    def __init__(self, data_path="data"):
        dataset = "FB13S"
        path = osp.join(data_path, dataset)
        super(FB13SDatset, self).__init__(path, dataset)



from typing import Optional, Tuple

import torch

from cogdl.utils.graph_utils import symmetric_normalization, row_normalization


class DropFeatures(torch.nn.Module):
    def __init__(self, drop_rate):
        super(DropFeatures, self).__init__()
        self.drop_rate = drop_rate

    def forward(self, x):
        return dropout_features(x, self.drop_rate, training=self.training)


class DropEdge(torch.nn.Module):
    def __init__(self, drop_rate: float = 0.5, renorm: Optional[str] = "sym"):
        super(DropEdge, self).__init__()
        self.drop_rate = drop_rate
        self.renorm = renorm

    def forward(self, edge_index, edge_weight=None):
        return dropout_adj(edge_index, edge_weight, self.drop_rate, self.renorm, self.training)


class DropNode(torch.nn.Module):
    def __init__(self, drop_rate=0.5):
        super(DropNode, self).__init__()
        self.drop_rate = drop_rate

    def forward(self, x):
        return drop_node(x, self.drop_rate, self.training)


def filter_adj(row, col, edge_attr, mask):
    return (row[mask], col[mask]), None if edge_attr is None else edge_attr[mask]


def dropout_adj(
    edge_index: Tuple,
    edge_weight: Optional[torch.Tensor] = None,
    drop_rate: float = 0.5,
    renorm: Optional[str] = "sym",
    training: bool = False,
):
    if not training or drop_rate == 0:
        if edge_weight is None:
            edge_weight = torch.ones(edge_index[0].shape[0], device=edge_index[0].device)
        return edge_index, edge_weight

    if drop_rate < 0.0 or drop_rate > 1.0:
        raise ValueError("Dropout probability has to be between 0 and 1, " "but got {}".format(drop_rate))

    row, col = edge_index
    num_nodes = int(max(row.max(), col.max())) + 1
    self_loop = row == col
    mask = torch.full((row.shape[0],), 1 - drop_rate, dtype=torch.float, device=row.device)
    mask = torch.bernoulli(mask).to(torch.bool)
    mask = self_loop | mask
    edge_index, edge_weight = filter_adj(row, col, edge_weight, mask)
    if renorm == "sym":
        edge_weight = symmetric_normalization(num_nodes, edge_index[0], edge_index[1])
    elif renorm == "row":
        edge_weight = row_normalization(num_nodes, edge_index[0], edge_index[1])
    return edge_index, edge_weight


def dropout_features(x: torch.Tensor, droprate: float, training=True):
    n = x.shape[1]
    drop_rates = torch.ones(n, device=x.device) * droprate
    if training:
        masks = torch.bernoulli(1.0 - drop_rates).view(1, -1).expand_as(x)
        masks = masks.to(x.device)
        masks = masks.to(x.device)
        x = masks * x
    return x


def drop_node(x, drop_rate=0.5, training=True):
    n = x.shape[0]
    drop_rates = torch.ones(n) * drop_rate
    if training:
        masks = torch.bernoulli(1.0 - drop_rates).unsqueeze(1)
        x = masks.to(x.device) * x
        x = x / drop_rate
    return x



from typing import Union, Callable
from abc import abstractmethod
import torch
from cogdl.wrappers.tools.wrapper_utils import merge_batch_indexes
from cogdl.utils.evaluator import setup_evaluator, Accuracy, MultiLabelMicroF1


class ModelWrapper(torch.nn.Module):
    @staticmethod
    def add_args(parser):
        pass

    def __init__(self):
        super(ModelWrapper, self).__init__()
        self.__model_keys__ = None
        self._loss_func = None
        self._evaluator = None
        self._evaluator_metric = None
        self.__record__ = dict()

    def forward(self):
        pass

    def pre_stage(self, stage, data_w):
        pass

    def post_stage(self, stage, data_w):
        pass

    def train_step(self, subgraph):
        pass

    def val_step(self, subgraph):
        pass

    def test_step(self, subgraph):
        pass

    def evaluate(self, pred: torch.Tensor, labels: torch.Tensor, metric: Union[str, Callable] = "auto"):
        """
        method: str or callable function,
        """
        pred = pred.cpu()
        labels = labels.cpu()
        if self._evaluator is None:
            if metric == "auto":
                if len(labels.shape) > 1:
                    metric = "multilabel_microf1"
                    self._evaluator_metric = "micro_f1"
                else:
                    metric = "accuracy"
                    self._evaluator_metric = "acc"

            self._evaluator = setup_evaluator(metric)
            # self._evaluator_metric = metric
        return self._evaluator(pred, labels)

    @abstractmethod
    def setup_optimizer(self):
        raise NotImplementedError

    def set_early_stopping(self):
        """
        Return:
            1. `str`, the monitoring metric
            2. tuple(`str`, `str`), that is, (the monitoring metric, `small` or `big`). The second parameter means,
                `the smaller, the better` or `the bigger, the better`
        """
        return "val_metric", ">"

    def on_train_step(self, *args, **kwargs):
        return self.train_step(*args, **kwargs)

    def on_val_step(self, *args, **kwargs):
        out = self.val_step(*args, **kwargs)
        self.set_notes(out, "val")

    def on_test_step(self, *args, **kwargs):
        out = self.test_step(*args, **kwargs)
        self.set_notes(out, "test")

    def set_notes(self, out, prefix="val"):
        if isinstance(out, dict):
            for key, val in out.items():
                self.note(key, val)
        elif isinstance(out, tuple) or isinstance(out, list):
            for i, val in enumerate(out):
                self.note(f"{prefix}_{i}", val)

    def note(self, name: str, data, merge="mean"):
        """
        name: str
        data: Any
        """
        if name not in self.__record__:
            name = name.lower()
            self.__record__[name] = [data]
            # self.__record_merge__[name] = merge
        else:
            self.__record__[name].append(data)

    def collect_notes(self):
        if len(self.__record__) == 0:
            return None
        out = dict()
        for key, val in self.__record__.items():
            if key.endswith("_metric"):
                _val = self._evaluator.evaluate()
                if _val == 0:
                    _val = val[0]
            # elif isinstance(self._evaluator_metric, str) and key.endswith(self._evaluator_metric):
            #     _val = self._evaluator.evaluate()
            else:
                _val = merge_batch_indexes(val)
            out[key] = _val
        self.__record__ = dict()
        return out

    @property
    def default_loss_fn(self):
        if self._loss_func is None:
            raise RuntimeError(
                "`loss_fn` must be set for your ModelWrapper using `mw.default_loss_fn = your_loss_fn`.",
                f"Now self.loss_fn is {self._loss_fn}",
            )
        return self._loss_func

    @default_loss_fn.setter
    def default_loss_fn(self, loss_fn):
        self._loss_func = loss_fn

    @property
    def default_evaluator(self):
        return self._evaluator

    @default_evaluator.setter
    def default_evaluator(self, x):
        self._evaluator = x

    @property
    def device(self):
        # for k in self._model_key_:
        #     return next(getattr(self, k).parameters()).device
        return next(self.parameters()).device

    @property
    def evaluation_metric(self):
        return self._evaluator_metric

    def set_evaluation_metric(self):
        if isinstance(self._evaluator, MultiLabelMicroF1):
            self._evaluator_metric = "micro_f1"
        elif isinstance(self._evaluator, Accuracy):
            self._evaluator_metric = "acc"
        else:
            evaluation_metric = self.set_early_stopping()
            if not isinstance(evaluation_metric, str):
                evaluation_metric = evaluation_metric[0]
            if evaluation_metric.startswith("val"):
                evaluation_metric = evaluation_metric[3:]
            self._evaluator_metric = evaluation_metric

    def load_checkpoint(self, path):
        pass

    def save_checkpoint(self, path):
        pass

    def _find_model(self):
        models = []
        for k, v in self.__dict__.items():
            if isinstance(v, torch.nn.Module):
                models.append(k)
        self.__model_keys__ = models

    @property
    def wrapped_model(self):
        if hasattr(self, "model"):
            return getattr(self, "model")
        assert len(self._model_key_) == 1, f"{len(self._model_key_)} exists"
        return getattr(self, self._model_key_[0])

    @wrapped_model.setter
    def wrapped_model(self, model):
        if len(self._model_key_) == 0:
            self.__model_keys__ = [None]
        setattr(self, self._model_key_[0], model)

    @property
    def _model_key_(self):
        if self.__model_keys__ is None:
            self._find_model()
        return self.__model_keys__


class EmbeddingModelWrapper(ModelWrapper):
    def setup_optimizer(self):
        pass



from functools import reduce
from scipy.special import iv

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_sparse import spspmm, spmm

from cogdl.utils import get_degrees


# ==========
# Attention
# ==========


class NodeAttention(nn.Module):
    def __init__(self, in_feat):
        super(NodeAttention, self).__init__()
        self.p = nn.Linear(in_feat, 1)
        self.dropout = nn.Dropout(0.7)

    def forward(self, x, edge_index, edge_attr):
        device = x.device
        N, dim = x.shape
        diag_val = self.p(x)
        diag_val = torch.sigmoid(diag_val)
        self.dropout(diag_val)

        row, col = edge_index
        deg = get_degrees(row, col, N)
        deg_inv = deg.pow(-1)
        edge_attr_t = deg_inv[row] * edge_attr

        diag_ind = torch.LongTensor([range(N)] * 2).to(device)
        _, adj_mat_val = spspmm(edge_index, edge_attr_t, diag_ind, diag_val.view(-1), N, N, N, True)
        return edge_index, adj_mat_val


class EdgeAttention(nn.Module):
    def __init__(self, in_feat):
        super(EdgeAttention, self).__init__()
        self.p = nn.Linear(in_feat, 1)
        self.q = nn.Linear(in_feat, 1)

        self.dropout = nn.Dropout(0.5)

    def forward(self, x, edge_index, edge_attr):
        N, dim = x.shape

        row, col = edge_index
        deg = get_degrees(row, col, N)
        deg_inv_sqrt = deg.pow(-0.5)
        edge_attr_t = deg_inv_sqrt[row] * edge_attr * deg_inv_sqrt[col]

        p_val = F.relu(self.p(x))
        q_val = F.relu(self.q(x))

        p_val = self.dropout(p_val)
        q_val = self.dropout(q_val)

        p_adj_mat_val = edge_attr_t * p_val.view(-1)[edge_index[1]]
        q_adj_mat_val = edge_attr_t * q_val.view(-1)[edge_index[0]]
        return edge_index, p_adj_mat_val + q_adj_mat_val


class Identity(nn.Module):
    def __init__(self, in_feat):
        super(Identity, self).__init__()

    def forward(self, x, edge_index, edge_attr):
        return edge_index, edge_attr


class Gaussian(nn.Module):
    def __init__(self, in_feat):
        super(Gaussian, self).__init__()
        self.mu = 0.2
        self.theta = 1.0
        self.steps = 4

    def forward(self, x, edge_index, edge_attr):
        N = x.shape[0]
        row, col = edge_index
        deg = get_degrees(row, col, N)
        deg_inv = deg.pow(-1)
        adj = torch.sparse_coo_tensor(edge_index, deg_inv[row] * edge_attr, size=(N, N))
        identity = torch.sparse_coo_tensor([range(N)] * 2, torch.ones(N), size=(N, N)).to(x.device)
        laplacian = identity - adj

        t0 = identity
        t1 = laplacian - self.mu * identity
        t1 = t1.mm(t1.to_dense()).to_sparse()
        l_x = -0.5 * (t1 - identity)

        ivs = [iv(i, self.theta) for i in range(self.steps)]
        ivs[1:] = [(-1) ** i * 2 * x for i, x in enumerate(ivs[1:])]
        ivs = torch.tensor(ivs).to(x.device)
        result = [t0, l_x]
        for i in range(2, self.steps):
            result.append(2 * l_x.mm(result[i - 1].to_dense()).to_sparse().sub(result[i - 2]))

        result = [result[i] * ivs[i] for i in range(self.steps)]

        def fn(x, y):
            return x.add(y)

        res = reduce(fn, result)

        return res._indices(), res._values()


class PPR(nn.Module):
    def __init__(self, in_feat):
        super(PPR, self).__init__()
        self.alpha = 0.4
        self.steps = 4

    def forward(self, x, edge_index, edge_attr):
        row, col = edge_index
        deg = get_degrees(row, col, x.shape[0])
        deg_inv_sqrt = deg.pow(-0.5)
        edge_attr_t = deg_inv_sqrt[row] * edge_attr * deg_inv_sqrt[col]

        N = x.size(0)
        adj = torch.sparse_coo_tensor(edge_index, edge_attr_t, size=(N, N))

        theta = self.alpha * (1 - self.alpha)
        result = [theta * adj]

        for i in range(1, self.steps - 1):
            theta = theta * (1 - self.alpha)
            adj_ind, adj_val = spspmm(
                edge_index, edge_attr_t, result[i - 1]._indices(), result[i - 1]._values(), N, N, N, True
            )
            result.append(torch.sparse_coo_tensor(adj_ind, adj_val, size=(N, N)))

        identity = torch.sparse_coo_tensor([range(N)] * 2, torch.ones(N), size=(N, N)).to(x.device)
        result.append(self.alpha * identity)

        def fn(x, y):
            return x.add(y)

        res = reduce(fn, result)
        return res._indices(), res._values()


class HeatKernel(nn.Module):
    def __init__(self, in_feat):
        super(HeatKernel, self).__init__()
        self.t = nn.Parameter(
            torch.zeros(
                1,
            )
        )

    def forward(self, x, edge_index, edge_attr):
        row, col = edge_index
        deg = get_degrees(row, col, x.shape[0])
        deg_inv = deg.pow(-1)
        edge_attr_t = self.t * edge_attr * deg_inv[col] - self.t
        return edge_index, edge_attr_t.exp()


def act_attention(attn_type):
    if attn_type == "identity":
        return Identity
    elif attn_type == "node":
        return NodeAttention
    elif attn_type == "edge":
        return EdgeAttention
    elif attn_type == "ppr":
        return PPR
    elif attn_type == "heat":
        return HeatKernel
    elif attn_type == "gaussian":
        return Gaussian
    else:
        raise ValueError("no such attention type")


# ===============
# Normalization
# ===============


class NormIdentity(nn.Module):
    def __init__(self):
        super(NormIdentity, self).__init__()

    def forward(self, edge_index, edge_attr, N):
        return edge_attr


class RowUniform(nn.Module):
    def __init__(self):
        super(RowUniform, self).__init__()

    def forward(self, edge_index, edge_attr, N):
        device = edge_attr.device
        ones = torch.ones(N, 1, device=device)
        rownorm = 1.0 / spmm(edge_index, edge_attr, N, N, ones).view(-1)
        row = rownorm[edge_index[0]]
        edge_attr_t = row * edge_attr

        return edge_attr_t


class RowSoftmax(nn.Module):
    def __init__(self):
        super(RowSoftmax, self).__init__()

    def forward(self, edge_index, edge_attr, N):
        device = edge_attr.device
        edge_attr_t = torch.exp(edge_attr)
        ones = torch.ones(N, 1, device=device)
        rownorm = 1.0 / spmm(edge_index, edge_attr_t, N, N, ones).view(-1)
        row = rownorm[edge_index[0]]
        edge_attr_t = row * edge_attr_t

        return edge_attr_t


class ColumnUniform(nn.Module):
    def __init__(self):
        super(ColumnUniform, self).__init__()

    def forward(self, edge_index, edge_attr, N):
        device = edge_attr.device
        ones = torch.ones(N, 1, device=device)
        rownorm = 1.0 / spmm(edge_index, edge_attr, N, N, ones).view(-1)
        col = rownorm[edge_index[1]]
        edge_attr_t = col * edge_attr

        return edge_attr_t


class SymmetryNorm(nn.Module):
    def __init__(self):
        super(SymmetryNorm, self).__init__()

    def forward(self, edge_index, edge_attr, N):
        device = edge_attr.device
        ones = torch.ones(N, 1, device=device)
        rownorm = spmm(edge_index, edge_attr, N, N, ones).view(-1).pow(-0.5)
        row = rownorm[edge_index[0]]
        col = rownorm[edge_index[1]]
        edge_attr_t = row * edge_attr * col

        return edge_attr_t


def act_normalization(norm_type):
    if norm_type == "identity":
        return NormIdentity
    elif norm_type == "row_uniform":
        return RowUniform
    elif norm_type == "row_softmax":
        return RowSoftmax
    elif norm_type == "col_uniform":
        return ColumnUniform
    elif norm_type == "symmetry":
        return SymmetryNorm
    else:
        raise ValueError("no such normalization type")


# ============
# activation
# ============


def act_map(act):
    if act == "linear":
        return lambda x: x
    elif act == "elu":
        return torch.nn.functional.elu
    elif act == "sigmoid":
        return torch.sigmoid
    elif act == "tanh":
        return torch.tanh
    elif act == "relu":
        return torch.nn.functional.relu
    elif act == "relu6":
        return torch.nn.functional.relu6
    elif act == "softplus":
        return torch.nn.functional.softplus
    elif act == "leaky_relu":
        return torch.nn.functional.leaky_relu
    else:
        raise Exception("wrong activate function")



from .. import DataWrapper
from cogdl.data import Graph


class FullBatchNodeClfDataWrapper(DataWrapper):
    def __init__(self, dataset):
        super(FullBatchNodeClfDataWrapper, self).__init__(dataset)
        self.dataset = dataset

    def train_wrapper(self) -> Graph:
        return self.dataset.data

    def val_wrapper(self):
        return self.dataset.data

    def test_wrapper(self):
        return self.dataset.data

    def pre_transform(self):
        self.dataset.data.add_remaining_self_loops()



import os.path as osp
import pickle

import numpy as np
import torch
from cogdl.data import Graph, Dataset
from cogdl.utils import download_url, untar


class GTNDataset(Dataset):
    r"""The network datasets "ACM", "DBLP" and "IMDB" from the
    `"Graph Transformer Networks"
    <https://arxiv.org/abs/1911.06455>`_ paper.

    Args:
        root (string): Root directory where the dataset should be saved.
        name (string): The name of the dataset (:obj:`"gtn-acm"`,
            :obj:`"gtn-dblp"`, :obj:`"gtn-imdb"`).
    """

    def __init__(self, root, name):
        self.name = name
        self.url = f"https://github.com/cenyk1230/gtn-data/blob/master/{name}.zip?raw=true"
        super(GTNDataset, self).__init__(root)
        self.data = torch.load(self.processed_paths[0])
        self.num_edge = len(self.data.adj)
        self.num_nodes = self.data.x.shape[0]

    @property
    def raw_file_names(self):
        names = ["edges.pkl", "labels.pkl", "node_features.pkl"]
        return names

    @property
    def processed_file_names(self):
        return ["data.pt"]

    @property
    def num_classes(self):
        return torch.max(self.data.train_target).item() + 1

    def read_gtn_data(self, folder):
        edges = pickle.load(open(osp.join(folder, "edges.pkl"), "rb"))
        labels = pickle.load(open(osp.join(folder, "labels.pkl"), "rb"))
        node_features = pickle.load(open(osp.join(folder, "node_features.pkl"), "rb"))

        data = Graph()
        data.x = torch.from_numpy(node_features).type(torch.FloatTensor)

        num_nodes = edges[0].shape[0]

        node_type = np.zeros((num_nodes), dtype=int)
        assert len(edges) == 4
        assert len(edges[0].nonzero()) == 2

        node_type[edges[0].nonzero()[0]] = 0
        node_type[edges[0].nonzero()[1]] = 1
        node_type[edges[1].nonzero()[0]] = 1
        node_type[edges[1].nonzero()[1]] = 0
        node_type[edges[2].nonzero()[0]] = 0
        node_type[edges[2].nonzero()[1]] = 2
        node_type[edges[3].nonzero()[0]] = 2
        node_type[edges[3].nonzero()[1]] = 0

        print(node_type)
        data.pos = torch.from_numpy(node_type)

        edge_list = []
        for i, edge in enumerate(edges):
            edge_tmp = torch.from_numpy(np.vstack((edge.nonzero()[0], edge.nonzero()[1]))).type(torch.LongTensor)
            edge_list.append(edge_tmp)
        data.edge_index = torch.cat(edge_list, 1)

        A = []
        for i, edge in enumerate(edges):
            edge_tmp = torch.from_numpy(np.vstack((edge.nonzero()[0], edge.nonzero()[1]))).type(torch.LongTensor)
            value_tmp = torch.ones(edge_tmp.shape[1]).type(torch.FloatTensor)
            A.append((edge_tmp, value_tmp))
        edge_tmp = torch.stack((torch.arange(0, num_nodes), torch.arange(0, num_nodes))).type(torch.LongTensor)
        value_tmp = torch.ones(num_nodes).type(torch.FloatTensor)
        A.append((edge_tmp, value_tmp))
        data.adj = A

        data.train_node = torch.from_numpy(np.array(labels[0])[:, 0]).type(torch.LongTensor)
        data.train_target = torch.from_numpy(np.array(labels[0])[:, 1]).type(torch.LongTensor)
        data.valid_node = torch.from_numpy(np.array(labels[1])[:, 0]).type(torch.LongTensor)
        data.valid_target = torch.from_numpy(np.array(labels[1])[:, 1]).type(torch.LongTensor)
        data.test_node = torch.from_numpy(np.array(labels[2])[:, 0]).type(torch.LongTensor)
        data.test_target = torch.from_numpy(np.array(labels[2])[:, 1]).type(torch.LongTensor)

        y = np.zeros((num_nodes), dtype=int)
        x_index = torch.cat((data.train_node, data.valid_node, data.test_node))
        y_index = torch.cat((data.train_target, data.valid_target, data.test_target))
        y[x_index.numpy()] = y_index.numpy()
        data.y = torch.from_numpy(y)
        self.data = data

    def get(self, idx):
        assert idx == 0
        return self.data

    def apply_to_device(self, device):
        self.data.x = self.data.x.to(device)
        self.data.y = self.data.y.to(device)

        self.data.train_node = self.data.train_node.to(device)
        self.data.valid_node = self.data.valid_node.to(device)
        self.data.test_node = self.data.test_node.to(device)

        self.data.train_target = self.data.train_target.to(device)
        self.data.valid_target = self.data.valid_target.to(device)
        self.data.test_target = self.data.test_target.to(device)

        new_adj = []
        for (t1, t2) in self.data.adj:
            new_adj.append((t1.to(device), t2.to(device)))
        self.data.adj = new_adj

    def download(self):
        download_url(self.url, self.raw_dir, name=self.name + ".zip")
        untar(self.raw_dir, self.name + ".zip")

    def process(self):
        self.read_gtn_data(self.raw_dir)
        torch.save(self.data, self.processed_paths[0])

    def __repr__(self):
        return "{}".format(self.name)


class ACM_GTNDataset(GTNDataset):
    def __init__(self, data_path="data"):
        dataset = "gtn-acm"
        path = osp.join(data_path, dataset)
        super(ACM_GTNDataset, self).__init__(path, dataset)


class DBLP_GTNDataset(GTNDataset):
    def __init__(self, data_path="data"):
        dataset = "gtn-dblp"
        path = osp.join(data_path, dataset)
        super(DBLP_GTNDataset, self).__init__(path, dataset)


class IMDB_GTNDataset(GTNDataset):
    def __init__(self, data_path="data"):
        dataset = "gtn-imdb"
        path = osp.join(data_path, dataset)
        super(IMDB_GTNDataset, self).__init__(path, dataset)



from .heterogeneous_embedding_dw import HeterogeneousEmbeddingDataWrapper
from .heterogeneous_gnn_dw import HeterogeneousGNNDataWrapper
from .multiplex_embedding_dw import MultiplexEmbeddingDataWrapper



"""A wrapper class for optimizer """
import numpy as np
import torch.nn as nn


class NoamOptimizer(nn.Module):
    """A simple wrapper class for learning rate scheduling"""

    def __init__(self, optimizer, d_model, n_warmup_steps, init_lr=None):
        super(NoamOptimizer, self).__init__()
        self._optimizer = optimizer
        self.param_groups = optimizer.param_groups
        self.n_warmup_steps = n_warmup_steps
        self.n_current_steps = 0
        self.init_lr = np.power(d_model, -0.5) if init_lr is None else init_lr / np.power(self.n_warmup_steps, -0.5)

    def step(self):
        """Step with the inner optimizer"""
        self._update_learning_rate()
        self._optimizer.step()

    def zero_grad(self):
        """Zero out the gradients by the inner optimizer"""
        self._optimizer.zero_grad()

    def _get_lr_scale(self):
        return np.min(
            [np.power(self.n_current_steps, -0.5), np.power(self.n_warmup_steps, -1.5) * self.n_current_steps]
        )

    def _update_learning_rate(self):
        """Learning rate scheduling per step"""

        self.n_current_steps += 1
        lr = self.init_lr * self._get_lr_scale()

        for param_group in self._optimizer.param_groups:
            param_group["lr"] = lr


class LinearOptimizer(nn.Module):
    """A simple wrapper class for learning rate scheduling"""

    def __init__(self, optimizer, n_warmup_steps, n_training_steps, init_lr=0.001):
        super(LinearOptimizer, self).__init__()
        self._optimizer = optimizer
        self.param_groups = optimizer.param_groups
        self.n_warmup_steps = n_warmup_steps
        self.n_current_steps = 0
        self.n_training_steps = n_training_steps
        self.init_lr = init_lr

    def step(self):
        """Step with the inner optimizer"""
        self._update_learning_rate()
        self._optimizer.step()

    def zero_grad(self):
        """Zero out the gradients by the inner optimizer"""
        self._optimizer.zero_grad()

    def _get_lr_scale(self):
        if self.n_current_steps < self.n_warmup_steps:
            return float(self.n_current_steps) / float(max(1, self.n_warmup_steps))
        return max(
            0.0,
            float(self.n_training_steps - self.n_current_steps)
            / float(max(1, self.n_training_steps - self.n_warmup_steps)),
        )

    def _update_learning_rate(self):
        """Learning rate scheduling per step"""

        self.n_current_steps += 1
        lr = self.init_lr * self._get_lr_scale()

        for param_group in self._optimizer.param_groups:
            param_group["lr"] = lr



import torch.nn as nn
import torch.nn.functional as F

from cogdl.layers import MixHopLayer

from .. import BaseModel


class MixHop(BaseModel):
    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--num-classes", type=int)
        parser.add_argument("--dropout", type=float, default=0.7)
        parser.add_argument("--layer1-pows", type=int, nargs="+", default=[200, 200, 200])
        parser.add_argument("--layer2-pows", type=int, nargs="+", default=[20, 20, 20])
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.num_classes,
            args.dropout,
            args.layer1_pows,
            args.layer2_pows,
        )

    def __init__(self, num_features, num_classes, dropout, layer1_pows, layer2_pows):
        super(MixHop, self).__init__()

        self.dropout = dropout

        self.num_features = num_features
        self.num_classes = num_classes
        self.dropout = dropout
        layer_pows = [layer1_pows, layer2_pows]

        shapes = [num_features] + [sum(layer1_pows), sum(layer2_pows)]

        self.mixhops = nn.ModuleList(
            [MixHopLayer(shapes[layer], [0, 1, 2], layer_pows[layer]) for layer in range(len(layer_pows))]
        )
        self.fc = nn.Linear(shapes[-1], num_classes)

    def forward(self, graph):
        x = graph.x
        for mixhop in self.mixhops:
            x = F.relu(mixhop(graph, x))
            x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.fc(x)
        return x

    def predict(self, data):
        return self.forward(data)



import copy
import itertools
import os
import inspect
from collections import defaultdict, namedtuple
import warnings

import torch
import torch.nn as nn
import optuna
from tabulate import tabulate

from cogdl.utils import set_random_seed, tabulate_results
from cogdl.configs import BEST_CONFIGS
from cogdl.data import Dataset
from cogdl.models import build_model
from cogdl.datasets import build_dataset
from cogdl.wrappers import fetch_model_wrapper, fetch_data_wrapper
from cogdl.options import get_default_args
from cogdl.trainer import Trainer


class AutoML(object):
    """
    Args:
        search_space: function to obtain hyper-parameters to search
    """

    def __init__(self, args):
        self.search_space = args.search_space
        self.metric = args.metric if hasattr(args, "metric") else None
        self.n_trials = args.n_trials if hasattr(args, "n_trials") else 3
        self.best_value = None
        self.best_params = None
        self.default_params = args

    def _objective(self, trials):
        params = copy.deepcopy(self.default_params)
        cur_params = self.search_space(trials)
        print(cur_params)
        for key, value in cur_params.items():
            params.__setattr__(key, value)
        result_dict = raw_experiment(args=params)
        result_list = list(result_dict.values())[0]
        item = result_list[0]
        key = self.metric
        if key is None:
            for _key in item.keys():
                if "Val" in _key or "val" in _key:
                    key = _key
                    break
            if key is None:
                raise KeyError("Unable to find validation metrics")
        val = [result[key] for result in result_list]
        mean = sum(val) / len(val)

        if self.best_value is None or mean > self.best_value:
            self.best_value = mean
            self.best_params = cur_params
            self.best_results = result_list

        return mean

    def run(self):
        study = optuna.create_study(direction="maximize")
        study.optimize(self._objective, n_trials=self.n_trials, n_jobs=1)
        # fig1 = optuna.visualization.plot_optimization_history(study)
        # fig1.show()
        # fig2 = optuna.visualization.plot_slice(study)
        # fig2.show()
        # fig3 = optuna.visualization.plot_param_importances(study)
        # fig3.show()
        print(study.best_params)
        return self.best_results


def set_best_config(args):
    if args.model not in BEST_CONFIGS:
        return args
    configs = BEST_CONFIGS[args.model]
    for key, value in configs["general"].items():
        args.__setattr__(key, value)
    if args.dataset not in configs:
        return args
    for key, value in configs[args.dataset].items():
        args.__setattr__(key, value)
    return args


def train(args):  # noqa: C901
    if isinstance(args.dataset, list):
        args.dataset = args.dataset[0]
    if isinstance(args.model, list):
        args.model = args.model[0]
    if isinstance(args.seed, list):
        args.seed = args.seed[0]
    if isinstance(args.split, list):
        args.split = args.split[0]
    set_random_seed(args.seed)

    model_name = args.model if isinstance(args.model, str) else args.model.model_name
    dw_name = args.dw if isinstance(args.dw, str) else args.dw.__name__
    mw_name = args.mw if isinstance(args.mw, str) else args.mw.__name__

    print(
        f""" 
|-------------------------------------{'-' * (len(str(args.dataset)) + len(model_name) + len(dw_name) + len(mw_name))}|
    *** Running (`{args.dataset}`, `{model_name}`, `{dw_name}`, `{mw_name}`)
|-------------------------------------{'-' * (len(str(args.dataset)) + len(model_name) + len(dw_name) + len(mw_name))}|"""
    )

    if getattr(args, "use_best_config", False):
        args = set_best_config(args)

    # setup dataset and specify `num_features` and `num_classes` for model
    if isinstance(args.dataset, Dataset):
        dataset = args.dataset
    else:
        dataset = build_dataset(args)

    mw_class = fetch_model_wrapper(args.mw)
    dw_class = fetch_data_wrapper(args.dw)

    if mw_class is None:
        raise NotImplementedError("`model wrapper(--mw)` must be specified.")

    if dw_class is None:
        raise NotImplementedError("`data wrapper(--dw)` must be specified.")

    data_wrapper_args = dict()
    model_wrapper_args = dict()
    # unworthy code: share `args` between model and dataset_wrapper
    for key in inspect.signature(dw_class).parameters.keys():
        if hasattr(args, key) and key != "dataset":
            data_wrapper_args[key] = getattr(args, key)
    # unworthy code: share `args` between model and model_wrapper
    for key in inspect.signature(mw_class).parameters.keys():
        if hasattr(args, key) and key != "model":
            model_wrapper_args[key] = getattr(args, key)

    # setup data_wrapper
    dataset_wrapper = dw_class(dataset, **data_wrapper_args)

    args.num_features = dataset.num_features
    if hasattr(dataset, "num_nodes"):
        args.num_nodes = dataset.num_nodes
    if hasattr(dataset, "num_edges"):
        args.num_edges = dataset.num_edges
    if hasattr(dataset, "num_edge"):
        args.num_edge = dataset.num_edge
    if hasattr(dataset, "max_graph_size"):
        args.max_graph_size = dataset.max_graph_size
    if hasattr(dataset, "edge_attr_size"):
        args.edge_attr_size = dataset.edge_attr_size
    else:
        args.edge_attr_size = [0]
    if hasattr(args, "unsup") and args.unsup:
        args.num_classes = args.hidden_size
    else:
        args.num_classes = dataset.num_classes
    if hasattr(dataset.data, "edge_attr") and dataset.data.edge_attr is not None:
        args.num_entities = len(torch.unique(torch.stack(dataset.data.edge_index)))
        args.num_rels = len(torch.unique(dataset.data.edge_attr))

    # setup model
    if isinstance(args.model, nn.Module):
        model = args.model
    else:
        model = build_model(args)
    # specify configs for optimizer
    optimizer_cfg = dict(
        lr=args.lr,
        weight_decay=args.weight_decay,
        n_warmup_steps=args.n_warmup_steps,
        epochs=args.epochs,
        batch_size=args.batch_size if hasattr(args, "batch_size") else 0,
    )

    if hasattr(args, "hidden_size"):
        optimizer_cfg["hidden_size"] = args.hidden_size

    # setup model_wrapper
    if isinstance(args.mw, str) and "embedding" in args.mw:
        model_wrapper = mw_class(model, **model_wrapper_args)
    else:
        model_wrapper = mw_class(model, optimizer_cfg, **model_wrapper_args)

    os.makedirs("./checkpoints", exist_ok=True)

    # setup controller
    trainer = Trainer(
        epochs=args.epochs,
        device_ids=args.devices,
        cpu=args.cpu,
        save_emb_path=args.save_emb_path,
        load_emb_path=args.load_emb_path,
        cpu_inference=args.cpu_inference,
        progress_bar=args.progress_bar,
        distributed_training=args.distributed,
        checkpoint_path=args.checkpoint_path,
        resume_training=args.resume_training,
        patience=args.patience,
        eval_step=args.eval_step,
        logger=args.logger,
        log_path=args.log_path,
        project=args.project,
        no_test=args.no_test,
        nstage=args.nstage,
        actnn=args.actnn,
    )

    # Go!!!
    result = trainer.run(model_wrapper, dataset_wrapper)

    return result


def gen_variants(**items):
    Variant = namedtuple("Variant", items.keys())
    return itertools.starmap(Variant, itertools.product(*items.values()))


def variant_args_generator(args, variants):
    """Form variants as group with size of num_workers"""
    for variant in variants:
        args.dataset, args.model, args.seed, args.split = variant
        yield copy.deepcopy(args)


def output_results(results_dict, tablefmt="github"):
    variant = list(results_dict.keys())[0]
    col_names = ["Variant"] + list(results_dict[variant][-1].keys())
    tab_data = tabulate_results(results_dict)

    print(tabulate(tab_data, headers=col_names, tablefmt=tablefmt))


def raw_experiment(args):
    variants = list(gen_variants(dataset=args.dataset, model=args.model, seed=args.seed, split=args.split))

    results_dict = defaultdict(list)
    results = [train(args) for args in variant_args_generator(args, variants)]
    for variant, result in zip(variants, results):
        results_dict[variant[:-2]].append(result)

    tablefmt = args.tablefmt if hasattr(args, "tablefmt") else "github"
    output_results(results_dict, tablefmt)

    return results_dict


def auto_experiment(args):
    variants = list(gen_variants(dataset=args.dataset, model=args.model))

    results_dict = defaultdict(list)
    for variant in variants:
        args.model = [variant.model]
        args.dataset = [variant.dataset]
        tool = AutoML(args)
        results_dict[variant[:]] = tool.run()

    tablefmt = args.tablefmt if hasattr(args, "tablefmt") else "github"
    print("\nFinal results:\n")
    output_results(results_dict, tablefmt)

    return results_dict


def default_search_space(trial):
    return {
        "dropout": trial.suggest_uniform("dropout", 0.2, 0.6),  # intra-layer
        "norm": trial.suggest_categorical("norm", ["batchnorm", "layernorm"]),
        "activation": trial.suggest_categorical("activation", ["relu", "gelu"]),
        "layers_type": trial.suggest_categorical("layers_type", ["gcn", "gat", "grand", "gcnii", "drgat"]),
        "residual": trial.suggest_categorical("residual", [True, False]),  # inter-layer
        "num_layers": trial.suggest_categorical("num_layers", [2, 4, 8]),
        "lr": trial.suggest_categorical("lr", [1e-3, 5e-3, 1e-2]),  # config
        "hidden_size": trial.suggest_categorical("hidden_size", [32, 64, 128]),
        # "optimizer": trial.suggest_categorical("optimizer", ["sgd", "adam"]),
        # "epochs": trial.suggest_categorical("epochs", [500, 1000, 1500]),
        "weight_decay": trial.suggest_categorical("weight_decay", [0, 1e-5, 1e-4]),
    }


def experiment(dataset, model=None, **kwargs):
    if model is None:
        model = "autognn"
    if isinstance(dataset, str) or isinstance(dataset, Dataset):
        dataset = [dataset]
    if isinstance(model, str) or isinstance(model, nn.Module):
        model = [model]
    if "args" not in kwargs:
        args = get_default_args(dataset=[str(x) for x in dataset], model=[str(x) for x in model], **kwargs)
    else:
        args = kwargs["args"]
        for key, value in kwargs.items():
            if key != "args":
                args.__setattr__(key, value)
    if isinstance(model[0], nn.Module):
        args.model = [x.model_name for x in model]
    print(args)
    args.dataset = dataset
    args.model = model

    if args.max_epoch is not None:
        warnings.warn("The max_epoch is deprecated and will be removed in the future, please use epochs instead!")
        args.epochs = args.max_epoch

    if len(model) == 1 and isinstance(model[0], str) and model[0] == "autognn":
        if not hasattr(args, "search_space"):
            args.search_space = default_search_space
        if not hasattr(args, "seed"):
            args.seed = [1, 2]
        if not hasattr(args, "n_trials"):
            args.n_trials = 20

    if hasattr(args, "search_space"):
        return auto_experiment(args)

    return raw_experiment(args)



import tvm
from tvm import te
from tvm.topi.util import get_const_tuple


def vanilla_spmm_csr_x86(SrcFeat,
                         Adj_indptr,
                         Adj_indices,
                         Adj_vals,
                         num_feat_partitions=1):
    """Compute sparse-dense matrix multiplication of Adj and SrcFeat on x86.
    This implementation applies feature dimension partitioning, which requires transforming the layout of SrcFeat.

    Parameters
    ----------
    SrcFeat : tvm.te.Tensor
        2-D with shape [num_src_vertices, feat_len]

    Adj_indptr : tvm.te.Tensor
        1-D with shape [num_dst_vertices + 1] (CSR)

    Adj_indices : tvm.te.Tensor
        1-D with shape [nnz] (CSR)

    Adj_vals : tvm.te.Tensor
        1-D with shape [nnz] (CSR)

    num_feat_partitions : int
        Doing feature dimension tiling

    Returns
    -------
    Out : tvm.te.Tensor
        2-D with shape [num_dst_vertices, feat_len]
    """
    assert Adj_indices.shape[0].value == Adj_vals.shape[0].value
    num_src_vertices, feat_len = get_const_tuple(SrcFeat.shape)
    num_dst_vertices = Adj_indptr.shape[0].value - 1
    oshape = (num_dst_vertices, feat_len)

    feat_len_per_partition = feat_len // num_feat_partitions  # we assume feat_len % num_feat_partitions = 0

    ReshapedSrcFeat = te.compute((num_feat_partitions, num_src_vertices, feat_len_per_partition), \
        lambda fo, nn, fi: SrcFeat[nn, fo * feat_len_per_partition + fi], name='ReshapedSrcFeat')

    def msgfunc(fo, row, fi):
        row_start = Adj_indptr[row]
        row_end = Adj_indptr[row + 1]
        row_num_elems = row_end - row_start
        elem_idx = te.reduce_axis((0, row_num_elems), name="elem_idx")
        adj_val = Adj_vals[row_start + elem_idx]
        feat_val = ReshapedSrcFeat[fo, Adj_indices[row_start + elem_idx], fi]
        return te.sum(adj_val * feat_val, axis=elem_idx)

    ReshapedOut = te.compute((num_feat_partitions, num_dst_vertices, feat_len_per_partition),
        msgfunc, name='ReshapedOut')

    Out = te.compute(oshape, \
        lambda nn, ff: ReshapedOut[ff // feat_len_per_partition, nn, ff % feat_len_per_partition], \
        name='Out')

    return Out


def schedule_vanilla_spmm_csr_x86(Out):
    s = te.create_schedule([Out.op])

    ReshapedOut = Out.op.input_tensors[0]
    ReshapedSrcFeat = ReshapedOut.op.input_tensors[3]

    # Reorder
    RO = ReshapedOut
    s[RO.op].reorder(RO.op.axis[0], RO.op.axis[1], RO.op.reduce_axis[0], RO.op.axis[2])

    # Parallelize the rows of the sparse matrix
    s[ReshapedSrcFeat.op].parallel(ReshapedSrcFeat.op.axis[1])
    s[ReshapedOut.op].parallel(ReshapedOut.op.axis[1])
    s[Out.op].parallel(Out.op.axis[0])

    return s


def vanilla_spmm_csr_cuda(SrcFeat,
                          Adj_indptr,
                          Adj_indices,
                          Adj_vals):
    """Compute sparse-dense matrix multiplication of Adj and SrcFeat on cuda.
    This implementation does not transform the layout of SrcFeat.

    Parameters
    ----------
    SrcFeat : tvm.te.Tensor
        2-D with shape [num_src_vertices, feat_len]

    Adj_indptr : tvm.te.Tensor
        1-D with shape [num_dst_vertices + 1] (CSR)

    Adj_indices : tvm.te.Tensor
        1-D with shape [nnz] (CSR)

    Adj_vals : tvm.te.Tensor
        1-D with shape [nnz] (CSR)

    Returns
    -------
    Out : tvm.te.Tensor
        2-D with shape [num_dst_vertices, feat_len]
    """
    assert Adj_indices.shape[0].value == Adj_vals.shape[0].value
    num_src_vertices, feat_len = get_const_tuple(SrcFeat.shape)
    num_dst_vertices = Adj_indptr.shape[0].value - 1
    oshape = (num_dst_vertices, feat_len)

    def msgfunc(row, ff):
        row_start = Adj_indptr[row]
        row_end = Adj_indptr[row + 1]
        row_num_elems = row_end - row_start
        elem_idx = te.reduce_axis((0, row_num_elems), name="elem_idx")
        adj_val = Adj_vals[row_start + elem_idx]
        feat_val = SrcFeat[Adj_indices[row_start + elem_idx], ff]
        return te.sum(adj_val * feat_val, axis=elem_idx)

    Out = te.compute(oshape, msgfunc, name='Out')

    return Out


def schedule_vanilla_spmm_csr_cuda(Out,
                                   num_cuda_blocks=None,
                                   num_threads_per_cuda_block=None):
    s = te.create_schedule([Out.op])
    num_rows = Out.shape[0].value
    feat_len = Out.shape[1].value
    if num_cuda_blocks is None:
        num_cuda_blocks = num_rows
    if num_threads_per_cuda_block is None:
        num_threads_per_cuda_block = feat_len
    row_axis = Out.op.axis[0]
    feat_axis = Out.op.axis[1]
    row_outer, row_inner = s[Out.op].split(row_axis, nparts=num_cuda_blocks)
    feat_outer, feat_inner = s[Out.op].split(feat_axis, factor=num_threads_per_cuda_block)
    s[Out.op].reorder(feat_outer, row_outer, feat_inner, row_inner)
    s[Out.op].bind(feat_outer, te.thread_axis("blockIdx.y"))
    s[Out.op].bind(row_outer, te.thread_axis("blockIdx.x"))
    s[Out.op].bind(feat_inner, te.thread_axis("threadIdx.x"))
    return s


def vanilla_spmm_dds_x86(SrcFeat,
                         Adj_s1_pos,
                         Adj_s1_idx,
                         Adj_vals,
                         d1_size,
                         d2_size,
                         num_feat_partitions=1):
    """Compute sparse-dense matrix multiplication of Adj and SrcFeat on x86.
    This implementation applies both feature dimension partitioning and 1D graph partitioning.
    1D graph partitioning transforms the csr Adj matrix into dense-dense-sparse (DDS) format.

    Parameters
    ----------
    SrcFeat : tvm.te.Tensor
        2-D with shape [num_src_vertices, feat_len]

    Adj_s1_pos : tvm.te.Tensor
        1-D with shape [d1_size * d2_size] (DDS)

    Adj_s1_idx : tvm.te.Tensor
        1-D with shape [nnz] (DDS)

    Adj_vals : tvm.te.Tensor
        1-D with shape [nnz] (DDS)

    d1_size : int
        Number of src vertex partitions

    d2_size : int
        num_dst_vertices + 1

    num_feat_partitions : int
        Doing feature dimension tiling

    Returns
    -------
    Out : tvm.te.Tensor
        2-D with shape [num_dst_vertices, feat_len]
    """
    assert d1_size * d2_size == Adj_s1_pos.shape[0].value
    assert Adj_s1_idx.shape[0].value == Adj_vals.shape[0].value
    num_src_vertices, feat_len = get_const_tuple(SrcFeat.shape)
    num_src_vertex_partitions = d1_size
    num_dst_vertices = d2_size - 1
    oshape = (num_dst_vertices, feat_len)

    feat_len_per_partition = feat_len // num_feat_partitions  # we assume feat_len % num_feat_partitions = 0
    num_src_vertices_per_partition = (num_src_vertices + num_src_vertex_partitions - 1) // num_src_vertex_partitions

    ReshapedSrcFeat = te.compute((num_feat_partitions, num_src_vertices, feat_len_per_partition), \
        lambda fo, nn, fi: SrcFeat[nn, fo * feat_len_per_partition + fi], name='ReshapedSrcFeat')

    def msgfunc(fo, src_vertex_partition_idx, row, fi):
        row_start = Adj_s1_pos[src_vertex_partition_idx * d2_size + row]
        row_end = Adj_s1_pos[src_vertex_partition_idx * d2_size + row + 1]
        row_num_elems = row_end - row_start
        elem_idx = te.reduce_axis((0, row_num_elems), name="elem_idx")
        adj_val = Adj_vals[row_start + elem_idx]
        feat_val = ReshapedSrcFeat[fo, \
                                   Adj_s1_idx[row_start + elem_idx] + src_vertex_partition_idx * num_src_vertices_per_partition, \
                                   fi]
        return te.sum(adj_val * feat_val, axis=elem_idx)

    Intermediate = te.compute((num_feat_partitions, num_src_vertex_partitions, num_dst_vertices, feat_len_per_partition), \
        msgfunc, name='Intermediate')

    k = te.reduce_axis((0, num_src_vertex_partitions), name='src_vertex_partition_reduce')
    ReshapedOut = te.compute((num_feat_partitions, num_dst_vertices, feat_len_per_partition),
        lambda fo, nn, fi: te.sum(Intermediate[fo, k, nn, fi], axis=k), \
        name='ReshapedOut')

    Out = te.compute(oshape, \
        lambda nn, ff: ReshapedOut[ff // feat_len_per_partition, nn, ff % feat_len_per_partition], \
        name='Out')

    return Out


def schedule_vanilla_spmm_dds_x86(Out):
    s = te.create_schedule([Out.op])

    ReshapedOut = Out.op.input_tensors[0]
    Intermediate = ReshapedOut.op.input_tensors[0]
    ReshapedSrcFeat = Intermediate.op.input_tensors[3]

    I = Intermediate
    RO = ReshapedOut
    s[I.op].reorder(I.op.axis[0], I.op.axis[1], I.op.axis[2], I.op.reduce_axis[0], I.op.axis[3])
    s[RO.op].reorder(RO.op.axis[0], RO.op.reduce_axis[0], RO.op.axis[1], RO.op.axis[2])
    s[I.op].compute_at(s[RO], RO.op.reduce_axis[0])

    # Parallelize the rows of the sparse matrix
    s[ReshapedSrcFeat.op].parallel(ReshapedSrcFeat.op.axis[1])
    s[Intermediate.op].parallel(Intermediate.op.axis[2])
    s[ReshapedOut.op].parallel(ReshapedOut.op.axis[1])
    s[Out.op].parallel(Out.op.axis[0])

    return s



import torch
import torch.nn as nn

from actnn.layers import QLinear, QReLU, QBatchNorm1d, QDropout

from cogdl.utils import spmm, get_activation


class ActMul(torch.autograd.Function):
    @staticmethod
    def forward(self, x, c):
        self.c = c
        return x * c

    @staticmethod
    def backward(self, grad_output):
        grad_input = grad_output.clone()
        grad_input *= self.c
        return grad_input, None

def act_mul(x, c):
    return ActMul.apply(x, c)

class ActGINLayer(nn.Module):
    r"""Graph Isomorphism Network layer from paper `"How Powerful are Graph
    Neural Networks?" <https://arxiv.org/pdf/1810.00826.pdf>`__.

    .. math::
        h_i^{(l+1)} = f_\Theta \left((1 + \epsilon) h_i^{l} +
        \mathrm{sum}\left(\left\{h_j^{l}, j\in\mathcal{N}(i)
        \right\}\right)\right)

    Parameters
    ----------
    apply_func : callable layer function)
        layer or function applied to update node feature
    eps : float32, optional
        Initial `\epsilon` value.
    train_eps : bool, optional
        If True, `\epsilon` will be a learnable parameter.
    """

    def __init__(self, in_features, out_features, pooled_features, apply_func=None, eps=0, train_eps=True, dropout=0.0, activation="relu", norm=None):
        super(ActGINLayer, self).__init__()
        if train_eps:
            self.eps = torch.nn.Parameter(torch.FloatTensor([eps]))
        else:
            self.register_buffer("eps", torch.FloatTensor([eps]))
        self.apply_func = apply_func
        # self.linear = QLinear(in_features, out_features)
        self.linear_prediction = QLinear(out_features, pooled_features)
        if dropout > 0:
            self.dropout = QDropout(dropout)
        else:
            self.dropout = None
        if activation is not None:
            self.act = QReLU()
        else:
            self.act = None

        if norm is not None:
            if norm == "batchnorm":
                self.norm = QBatchNorm1d(out_features)
            else:
                raise NotImplementedError
        else:
            self.norm = None

    def forward(self, graph, x):
        out = (1 + self.eps) * x + spmm(graph, x, actnn=True)
        if self.apply_func is not None:
            out = self.apply_func(out)
        # out = self.linear(out)

        if self.norm is not None:
            out = self.norm(out)
        if self.act is not None:
            out = self.act(out)

        # if self.dropout is not None:
        #     out = self.dropout(out)

        device = x.device
        batchsize = int(torch.max(graph.batch)) + 1

        pooled = self.linear_prediction(out)
        hsize = pooled.shape[1]
        pooled2 = torch.zeros(batchsize, hsize).to(device)
        pooled2.scatter_add_(dim=0, index=graph.batch.view(-1, 1).repeat(1, hsize), src=pooled)
        if self.dropout is not None:
            pooled2 = self.dropout(pooled2)

        return out, pooled2



import torch
import torch.nn.functional as F
from cogdl.utils import spmm

from . import BaseLayer


class GINELayer(BaseLayer):
    r"""The modified GINConv operator from the `"Graph convolutions that can finally model local structure" paper
     <https://arxiv.org/pdf/2011.15069.pdf>`__.

    Parameters
    ----------
    apply_func : callable layer function)
        layer or function applied to update node feature
    eps : float32, optional
        Initial `\epsilon` value.
    train_eps : bool, optional
        If True, `\epsilon` will be a learnable parameter.
    """

    def __init__(self, apply_func=None, eps=0, train_eps=True):
        super(GINELayer, self).__init__()
        if train_eps:
            self.eps = torch.nn.Parameter(torch.FloatTensor([eps]))
        else:
            self.register_buffer("eps", torch.FloatTensor([eps]))
        self.apply_func = apply_func

    def forward(self, graph, x):
        # m = self.message(x[graph.edge_index[0]], graph.edge_attr)
        # out = self.aggregate(graph, m)
        out = spmm(graph, x)
        out += (1 + self.eps) * x
        if self.apply_func is not None:
            out = self.apply_func(out)
        return out

    def message(self, x, attr):
        return F.relu(x + attr)



import time

import networkx as nx
import numpy as np
import scipy.sparse as sp
from sklearn import preprocessing
import torch
import torch.nn as nn
import torch.nn.functional as F
from tqdm import tqdm

from .. import BaseModel


class DNGR_layer(nn.Module):
    def __init__(self, num_node, hidden_size1, hidden_size2):
        super(DNGR_layer, self).__init__()
        self.num_node = num_node
        self.hidden_size1 = hidden_size1
        self.hidden_size2 = hidden_size2

        self.encoder = nn.Sequential(
            nn.Linear(self.num_node, self.hidden_size1),
            nn.Tanh(),
            nn.Linear(self.hidden_size1, self.hidden_size2),
            nn.Tanh(),
        )
        self.decoder = nn.Sequential(
            nn.Linear(self.hidden_size2, self.hidden_size1),
            nn.Tanh(),
            nn.Linear(self.hidden_size1, self.num_node),
            nn.Tanh(),
        )

    def forward(self, x):
        encoded = self.encoder(x)
        decoded = self.decoder(encoded)
        return encoded, decoded


class DNGR(BaseModel):
    r"""The DNGR model from the `"Deep Neural Networks for Learning Graph Representations"
    <https://www.aaai.org/ocs/index.php/AAAI/AAAI16/paper/download/12423/11715>`_ paper

    Args:
        hidden_size1 (int) : The size of the first hidden layer.
        hidden_size2 (int) : The size of the second hidden layer.
        noise (float) : Denoise rate of DAE.
        alpha (float) : Parameter in DNGR.
        step (int) : The max step in random surfing.
        epochs (int) : The max epoches in training step.
        lr (float) : Learning rate in DNGR.
    """

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--hidden-size1", type=int, default=1000, help="Hidden size in first layer of Auto-Encoder")
        parser.add_argument("--hidden-size2", type=int, default=128, help="Hidden size in second layer of Auto-Encoder")
        parser.add_argument("--noise", type=float, default=0.2, help="denoise rate of DAE")
        parser.add_argument("--alpha", type=float, default=0.98, help="alhpa is a hyperparameter in DNGR")
        parser.add_argument("--step", type=int, default=10, help="step is a hyperparameter in DNGR")

        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.hidden_size1, args.hidden_size2, args.noise, args.alpha, args.step, args.epochs, args.lr, args.cpu
        )

    def __init__(self, hidden_size1, hidden_size2, noise, alpha, step, epochs, lr, cpu):
        super(DNGR, self).__init__()
        self.hidden_size1 = hidden_size1
        self.hidden_size2 = hidden_size2
        self.noise = noise
        self.alpha = alpha
        self.step = step
        self.epochs = epochs
        self.lr = lr
        self.cpu = cpu

    def scale_matrix(self, mat):
        mat = mat - np.diag(np.diag(mat))
        D_inv = np.diagflat(np.reciprocal(np.sum(mat, axis=0)))
        mat = np.dot(D_inv, mat)
        return mat

    def random_surfing(self, adj_matrix):
        # Random Surfing
        adj_matrix = self.scale_matrix(adj_matrix)
        P0 = np.eye(self.num_node, dtype="float32")
        M = np.zeros((self.num_node, self.num_node), dtype="float32")
        P = np.eye(self.num_node, dtype="float32")
        for i in range(0, self.step):
            P = self.alpha * np.dot(P, adj_matrix) + (1 - self.alpha) * P0
            M = M + P
        return M

    def get_ppmi_matrix(self, mat):
        # Get Positive Pairwise Mutual Information(PPMI) matrix
        mat = self.random_surfing(mat)
        M = self.scale_matrix(mat)
        col_s = np.sum(M, axis=0).reshape(1, self.num_node)
        row_s = np.sum(M, axis=1).reshape(self.num_node, 1)
        D = np.sum(col_s)
        rowcol_s = np.dot(row_s, col_s)
        PPMI = np.log(np.divide(D * M, rowcol_s))

        PPMI[np.isnan(PPMI)] = 0.0
        PPMI[np.isinf(PPMI)] = 0.0
        PPMI[np.isneginf(PPMI)] = 0.0
        PPMI[PPMI < 0] = 0.0
        return PPMI

    def get_denoised_matrix(self, mat):
        return mat * (np.random.random(mat.shape) > self.noise)

    def get_emb(self, matrix):
        ut, s, _ = sp.linalg.svds(matrix, self.hidden_size2)
        emb_matrix = ut * np.sqrt(s)
        emb_matrix = preprocessing.normalize(emb_matrix, "l2")
        return emb_matrix

    def forward(self, graph, return_dict=False):
        device = "cuda" if torch.cuda.is_available() and not self.cpu else "cpu"
        G = graph.to_networkx()
        self.num_node = G.number_of_nodes()
        A = nx.adjacency_matrix(G).todense()
        PPMI = self.get_ppmi_matrix(A)
        print("PPMI matrix compute done")
        # return self.get_emb(PPMI)

        input_mat = torch.from_numpy(self.get_denoised_matrix(PPMI).astype(np.float32))
        model = DNGR_layer(self.num_node, self.hidden_size1, self.hidden_size2)

        input_mat = input_mat.to(device)
        model = model.to(device)

        opt = torch.optim.Adam(model.parameters(), lr=self.lr)
        loss_func = nn.MSELoss()

        epoch_iter = tqdm(range(self.epochs))
        for epoch in epoch_iter:
            opt.zero_grad()
            encoded, decoded = model.forward(input_mat)
            Loss = loss_func(decoded, input_mat)
            Loss.backward()
            epoch_iter.set_description(f"Epoch: {epoch:03d},  Loss: {Loss:.8f}")
            opt.step()
        embeddings, _ = model.forward(input_mat)
        embeddings = embeddings.detach().cpu().numpy()

        if return_dict:
            features_matrix = dict()
            for vid, node in enumerate(G.nodes()):
                features_matrix[node] = embeddings[vid]
        else:
            features_matrix = np.zeros((graph.num_nodes, embeddings.shape[1]))
            nx_nodes = G.nodes()
            features_matrix[nx_nodes] = embeddings[np.arange(graph.num_nodes)]
        return features_matrix



from .graph_classification_dw import GraphClassificationDataWrapper
from .graph_embedding_dw import GraphEmbeddingDataWrapper
from .infograph_dw import InfoGraphDataWrapper
from .patchy_san_dw import PATCHY_SAN_DataWrapper



import os.path as osp
import os
import json
from cogdl.data import Dataset
from cogdl.utils import download_url, untar


url_map = {
    "l0fos": "https://cloud.tsinghua.edu.cn/f/cd6e3f3276c14e73a9f7/?dl=1",
    "aff30": "https://cloud.tsinghua.edu.cn/f/c37df92984214b9bad66/?dl=1",
    "arxivvenue": "https://cloud.tsinghua.edu.cn/f/5bd4597726514a119898/?dl=1",
}


class oagbert_dataset(Dataset):
    def __init__(self):
        self.url = url_map[self.name]
        path = osp.join("data", self.name)
        self.content = ["._SUCCESS"]
        self.candidates = []
        super(oagbert_dataset, self).__init__(path, self.name)

    def download(self):
        filename = self.name + ".zip"
        download_url(self.url, self.processed_dir, name=filename)
        untar(self.processed_dir, filename)
        print(f"downloaded to {self.processed_dir}")

    def process(self):
        pass

    def get_candidates(self):
        """
        return the list of candidates
        """
        with open("%s/._SUCCESS" % self.processed_dir) as f:
            for line in f:
                line = line.strip()
                self.candidates.append(line)
        return self.candidates

    def get_data(self):
        """
        Return all the file name
        Not expanded to the other two dataset yet
        """
        sample = {}
        for filename in os.listdir(self.processed_dir):
            if not filename.endswith(".jsonl"):
                continue
            sample_file = self.processed_dir + "/" + filename
            fin = open(sample_file, "r")
            file = []
            for line in fin:
                paper = json.loads(line.strip())
                file.append(paper)
            sample[filename] = file
        return sample

    @property
    def raw_file_names(self):
        return self.content

    def __len__(self):
        return 20

    @property
    def processed_file_names(self):
        return self.content


class l0fos(oagbert_dataset):
    def __init__(self):
        self.name = "l0fos"
        super(l0fos, self).__init__()


class aff30(oagbert_dataset):
    def __init__(self):
        self.name = "aff30"
        super(aff30, self).__init__()


class arxivvenue(oagbert_dataset):
    def __init__(self):
        self.name = "arxivvenue"
        super(arxivvenue, self).__init__()



import numpy as np
import torch
import torch.nn as nn
from cogdl.layers import FSGCNLayer
from cogdl.utils import spmm

from .. import BaseModel


class FSGCN(BaseModel):

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--num-classes", type=int)
        parser.add_argument("--num-layers", type=int, default=2)
        parser.add_argument("--hidden-size", type=int, default=64)
        parser.add_argument("--dropout", type=float, default=0.5)
        parser.add_argument("--residual", action="store_true")
        parser.add_argument("--norm", type=str, default=None)
        parser.add_argument("--activation", type=str, default="relu")
        parser.add_argument("--sample-size", type=int, default=[2], nargs="+", metavar="N")
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.hidden_size,
            args.num_classes,
            args.num_layers,
            args.dropout,
            args.activation,
            args.residual,
            args.norm,
            args.actnn,
            args.rp_ratio,
            args.sample_size,
        )

    def __init__(
        self,
        in_feats,
        hidden_size,
        out_feats,
        num_layers,
        dropout,
        activation="relu",
        residual=False,
        norm=None,
        actnn=False,
        rp_ratio=1,
        sample_size=[2],
    ):
        super(FSGCN, self).__init__()
        self.hist = []
        self.first_epoch = True
        if len(sample_size) == 1:
            self.sample_size = sample_size * (num_layers - 1)
        self.hidden_size = hidden_size
        self.out_feats = out_feats
        shapes = [in_feats] + [hidden_size] * (num_layers - 1) + [out_feats]
        Layer = FSGCNLayer
        self.layers = nn.ModuleList(
            [
                Layer(
                    shapes[i],
                    shapes[i + 1],
                    dropout=dropout if i != num_layers - 1 else 0,
                    residual=residual if i != num_layers - 1 else None,
                    norm=norm if i != num_layers - 1 else None,
                    activation=activation if i != num_layers - 1 else None,
                    rp_ratio=rp_ratio,
                )
                for i in range(num_layers)
            ]
        )
        self.num_layers = num_layers

    def forward(self, graph):
        graph.sym_norm()
        h = graph.x
        if not self.training:
            # self.hist = []
            for i in range(self.num_layers):
                # self.hist.append(spmm(graph, h))
                h = self.layers[i](graph, h)
            return h
        # with torch.no_grad():
        #     self.hist = []
        #     for i in range(self.num_layers):
        #         self.hist.append(spmm(graph, h.detach()))
        #         h = self.layers[i](graph, h)
        if self.first_epoch:
            for i in range(self.num_layers):
                self.hist.append(spmm(graph, h.detach()))
                h = self.layers[i](graph, h)
            self.first_epoch = False
        else:
            idx = torch.randint(self.hidden_size, (self.sample_size[0],)).to(h.device)
            # if np.random.random() < 0.5:
            #     idx = torch.arange(self.hidden_size // 2).to(h.device)
            # else:
            #     idx = torch.arange(start=self.hidden_size // 2, end = self.hidden_size).to(h.device)
            out_idx = torch.arange(self.out_feats).to(h.device)
            for i in range(self.num_layers):
                # cur_idx = torch.randint(self.hidden_size, (self.sample_size[i],)).to(h.device) if i < self.num_layers - 1 else out_idx
                cur_idx = idx if i < self.num_layers - 1 else out_idx
                if i > 0:
                    h = self.layers[i](graph, h, self.hist[i], idx, cur_idx)
                else:
                    h = self.layers[i](graph, h, self.hist[i], None, cur_idx)
                # if i < self.num_layers - 1:
                #     h = spmm(graph, h)
                #     self.hist[i + 1][:, cur_idx] = h.detach()
                    # with torch.no_grad():
                    #     self.hist[i + 1][:, cur_idx] = spmm(graph, h.detach())
                idx = cur_idx
        return h

    def predict(self, data):
        return self.forward(data)



import os
import random

import matplotlib.cm as cm
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import torch
from grave import plot_network, use_attributes
from tabulate import tabulate

from cogdl import oagbert
from cogdl.data import Graph
from cogdl.datasets import build_dataset_from_name, NodeDataset
from cogdl.models import build_model
from cogdl.options import get_default_args
from cogdl.experiments import train
from cogdl.datasets.rec_data import build_recommendation_data


class Pipeline(object):
    def __init__(self, app: str, **kwargs):
        self.app = app
        self.kwargs = kwargs

    def __call__(self, **kwargs):
        raise NotImplementedError


class DatasetPipeline(Pipeline):
    def __init__(self, app: str, **kwargs):
        super(DatasetPipeline, self).__init__(app, **kwargs)

    def __call__(self, dataset, **kwargs):
        if isinstance(dataset, str):
            dataset = [dataset]

        return self._call(dataset, **kwargs)


class DatasetStatsPipeline(DatasetPipeline):
    def __init__(self, app: str, **kwargs):
        super(DatasetStatsPipeline, self).__init__(app, **kwargs)

    def _call(self, dataset=[], **kwargs):
        if isinstance(dataset, str):
            dataset = [dataset]
        tab_data = []
        col_names = [
            "Dataset",
            "#nodes",
            "#edges",
            "#features",
            "#classes",
            "#labeled data",
        ]
        for name in dataset:
            dataset = build_dataset_from_name(name)
            data = dataset[0]

            tab_data.append(
                [
                    name,
                    data.x.shape[0],
                    data.edge_index[0].shape[0],
                    data.x.shape[1],
                    len(set(data.y.numpy())),
                    sum(data.train_mask.numpy()),
                ]
            )
        print(tabulate(tab_data, headers=col_names, tablefmt="psql"))

        return tab_data


class DatasetVisualPipeline(DatasetPipeline):
    def __init__(self, app: str, **kwargs):
        super(DatasetVisualPipeline, self).__init__(app, **kwargs)

    def _call(self, dataset="cora", seed=-1, depth=3, **kwargs):
        if isinstance(dataset, list):
            dataset = dataset[0]
        name = dataset
        dataset = build_dataset_from_name(name)
        data = dataset[0]

        G = nx.Graph()
        edge_index = torch.stack(data.edge_index)
        G.add_edges_from([tuple(edge_index[:, i].numpy()) for i in range(edge_index.shape[1])])

        if seed == -1:
            seed = random.choice(list(G.nodes()))
        q = [seed]
        node_set = set([seed])
        node_index = {seed: 0}
        max_index = 1
        for _ in range(depth):
            nq = []
            for x in q:
                for key in G[x].keys():
                    if key not in node_set:
                        nq.append(key)
                        node_set.add(key)
                        node_index[key] = node_index[x] + 1
            if len(nq) > 0:
                max_index += 1
            q = nq

        cmap = cm.rainbow(np.linspace(0.0, 1.0, max_index))

        for node, index in node_index.items():
            G.nodes[node]["color"] = cmap[index]
            G.nodes[node]["size"] = (max_index - index) * 50

        pic_file = f"{name}.png"
        plt.subplots()
        plot_network(G.subgraph(list(node_set)), node_style=use_attributes())
        plt.savefig(pic_file)
        print(f"Sampled ego network saved to {pic_file}")

        return q


class OAGBertInferencePipepline(Pipeline):
    def __init__(self, app: str, model: str, **kwargs):
        super(OAGBertInferencePipepline, self).__init__(app, model=model, **kwargs)

        load_weights = kwargs["load_weights"] if "load_weights" in kwargs else True
        self.tokenizer, self.bert_model = oagbert(model, load_weights=load_weights)

    def __call__(self, sequence, **kwargs):
        tokens = self.tokenizer(sequence, return_tensors="pt", padding=True)
        outputs = self.bert_model(**tokens)

        return outputs


class GenerateEmbeddingPipeline(Pipeline):
    def __init__(self, app: str, model: str, **kwargs):
        super(GenerateEmbeddingPipeline, self).__init__(app, model=model, **kwargs)

        self.kwargs = kwargs

        emb_models = [
            "prone",
            "netmf",
            "netsmf",
            "deepwalk",
            "line",
            "node2vec",
            "hope",
            "sdne",
            "grarep",
            "dngr",
            "spectral",
        ]
        gnn_models = ["dgi", "mvgrl", "grace", "unsup_graphsage"]

        if model in emb_models:
            self.method_type = "emb"
            args = get_default_args(dataset="blogcatalog", model=model, **kwargs)
        elif model in gnn_models:
            self.method_type = "gnn"
            args = get_default_args(dataset="cora", model=model, **kwargs)
        else:
            print("Please choose a model from ", emb_models, "or", gnn_models)
            exit(0)

        self.data_path = kwargs.get("data_path", "tmp_data.pt")
        self.num_features = kwargs.get("num_features", None)
        if self.num_features is not None:
            args.num_features = self.num_features
        elif self.method_type == "gnn":
            print("Please provide num_features for gnn model!")
            exit(0)

        args.model = args.model[0]
        self.args = args

    def __call__(self, edge_index, x=None, edge_weight=None):
        if self.method_type == "emb":
            if isinstance(edge_index, np.ndarray):
                edge_index = torch.from_numpy(edge_index)
            edge_index = (edge_index[:, 0], edge_index[:, 1])
            data = Graph(edge_index=edge_index, edge_weight=edge_weight)
            self.model = build_model(self.args)
            embeddings = self.model(data)
        elif self.method_type == "gnn":
            num_nodes = edge_index.max().item() + 1
            if x is None:
                print("No input node features, use random features instead.")
                x = np.random.randn(num_nodes, self.num_features)
            if isinstance(x, np.ndarray):
                x = torch.from_numpy(x).float()
            if isinstance(edge_index, np.ndarray):
                edge_index = torch.from_numpy(edge_index)
            edge_index = (edge_index[:, 0], edge_index[:, 1])
            data = Graph(x=x, edge_index=edge_index, edge_weight=edge_weight)
            torch.save(data, self.data_path)
            dataset = NodeDataset(path=self.data_path, scale_feat=False, metric="accuracy")
            self.args.dataset = dataset
            model = train(self.args)
            embeddings = model.embed(data.to(model.device))
            embeddings = embeddings.detach().cpu().numpy()

        return embeddings


class RecommendationPipepline(Pipeline):
    def __init__(self, app: str, model: str, **kwargs):
        super(RecommendationPipepline, self).__init__(app, model=model, **kwargs)

        if "data" in kwargs:
            data = kwargs["data"]
            val_data = test_data = data[-100:, :]
            data = build_recommendation_data("custom", data, val_data, test_data)
            self.data_path = kwargs.get("data_path", "tmp_data.pt")
            self.batch_size = kwargs.get("batch_size", 128)
            torch.save(data, self.data_path)
            self.dataset = NodeDataset(path=self.data_path, scale_feat=False)
        elif "dataset" in kwargs:
            dataset = kwargs.pop("dataset")
            self.dataset = build_dataset_from_name(dataset)
        else:
            print("Please provide recommendation data!")
            exit(0)

        self.batch_size = kwargs.get("batch_size", 2048)
        self.n_items = self.dataset[0].n_params["n_items"]

        args = get_default_args(task="recommendation", dataset="ali", model=model, **kwargs)
        args.model = args.model[0]

        # task = build_task(args, dataset=self.dataset)
        # task.train()

        # self.model = task.model
        self.model = build_model(args)
        self.model.eval()

        self.user_emb, self.item_emb = self.model.generate()

    def __call__(self, user_batch, **kwargs):
        user_batch = np.array(user_batch)
        user_batch = torch.from_numpy(user_batch).to(self.model.device)
        u_g_embeddings = self.user_emb[user_batch]

        # batch-item test
        n_item_batchs = self.n_items // self.batch_size + 1
        rate_batch = np.zeros(shape=(len(user_batch), self.n_items))

        i_count = 0
        for i_batch_id in range(n_item_batchs):
            i_start = i_batch_id * self.batch_size
            i_end = min((i_batch_id + 1) * self.batch_size, self.n_items)

            item_batch = torch.LongTensor(np.array(range(i_start, i_end))).view(i_end - i_start).to(self.model.device)
            i_g_embddings = self.item_emb[item_batch]

            i_rate_batch = self.model.rating(u_g_embeddings, i_g_embddings).detach().cpu()

            rate_batch[:, i_start:i_end] = i_rate_batch
            i_count += i_rate_batch.shape[1]

        topk = kwargs.get("topk", 10)
        results = {}
        for i in range(len(user_batch)):
            rate = list(zip(range(self.n_items), rate_batch[i]))
            rate.sort(key=lambda x: x[1], reverse=True)
            results[user_batch[i].item()] = [rate[j] for j in range(min(topk, len(rate)))]

        return results


SUPPORTED_APPS = {
    "dataset-stats": {"impl": DatasetStatsPipeline, "default": {"dataset": "cora"}},
    "dataset-visual": {"impl": DatasetVisualPipeline, "default": {"dataset": "cora"}},
    "oagbert": {"impl": OAGBertInferencePipepline, "default": {"model": "oagbert-v1"}},
    "generate-emb": {"impl": GenerateEmbeddingPipeline, "default": {"model": "prone"}},
    "recommendation": {"impl": RecommendationPipepline, "default": {"model": "lightgcn"}},
}


def check_app(app: str):
    if app in SUPPORTED_APPS:
        targeted_app = SUPPORTED_APPS[app]
        return targeted_app

    raise KeyError("Unknown app {}, available apps are {}".format(app, list(SUPPORTED_APPS.keys())))


def pipeline(app: str, **kwargs) -> Pipeline:
    targeted_app = check_app(app)
    task_class = targeted_app["impl"]
    default_args = targeted_app["default"].copy()
    default_args.update(kwargs)

    return task_class(app=app, **default_args)



import torch
import torch.nn as nn
from cogdl.utils import spmm


class MixHopLayer(nn.Module):
    def __init__(self, num_features, adj_pows, dim_per_pow):
        super(MixHopLayer, self).__init__()
        self.num_features = num_features
        self.adj_pows = adj_pows
        self.dim_per_pow = dim_per_pow
        self.total_dim = 0
        self.linears = torch.nn.ModuleList()
        for dim in dim_per_pow:
            self.linears.append(nn.Linear(num_features, dim))
            self.total_dim += dim
        # self.reset_parameters()

    def reset_parameters(self):
        for linear in self.linears:
            linear.reset_parameters()

    def adj_pow_x(self, graph, x, p):
        for _ in range(p):
            x = spmm(graph, x)
        return x

    def forward(self, graph, x):
        graph.sym_norm()
        output_list = []
        for p, linear in zip(self.adj_pows, self.linears):
            output = linear(self.adj_pow_x(graph, x, p))
            output_list.append(output)

        return torch.cat(output_list, dim=1)



import torch
from torch.utils.data import DataLoader

from .. import DataWrapper
from cogdl.models.nn.sagn import prepare_labels, prepare_feats


class SAGNDataWrapper(DataWrapper):
    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--batch-size", type=int, default=128)
        parser.add_argument("--label-nhop", type=int, default=3)
        parser.add_argument("--threshold", type=float, default=0.3)
        parser.add_argument("--nhop", type=int, default=3)
        # fmt: on

    def __init__(self, dataset, batch_size, label_nhop, threshold, nhop):
        super(SAGNDataWrapper, self).__init__(dataset)
        self.dataset = dataset
        self.batch_size = batch_size
        self.label_nhop = label_nhop
        self.nhop = nhop
        self.threshold = threshold

        self.label_emb, self.labels_with_pseudos, self.probs = None, None, None
        self.multihop_feats = None
        self.train_nid_with_pseudos = self.dataset.data.train_nid

        self.refresh_per_epoch("train")

    def train_wrapper(self):
        return DataLoader(self.train_nid_with_pseudos, batch_size=self.batch_size, shuffle=False)

    def val_wrapper(self):
        val_nid = self.dataset.data.val_nid
        return DataLoader(val_nid, batch_size=self.batch_size, shuffle=False)

    def test_wrapper(self):
        test_nid = self.dataset.data.test_nid
        return DataLoader(test_nid, batch_size=self.batch_size, shuffle=False)

    def post_stage_wrapper(self):
        data = self.dataset.data
        train_nid, val_nid, test_nid = data.train_nid, data.val_nid, data.test_nid
        all_nid = torch.cat([train_nid, val_nid, test_nid])
        return DataLoader(all_nid.numpy(), batch_size=self.batch_size, shuffle=False)

    def pre_stage_transform(self, batch):
        return self.train_transform(batch)

    def pre_transform(self):
        self.multihop_feats = prepare_feats(self.dataset, self.label_nhop)

    def train_transform(self, batch):
        batch_x = [x[batch] for x in self.multihop_feats]
        batch_x = torch.stack(batch_x)
        if self.label_emb is not None:
            batch_y_emb = self.label_emb[batch]
        else:
            batch_y_emb = None
        y = self.labels_with_pseudos[batch]
        return [batch_x, batch_y_emb, y]

    def val_transform(self, batch):
        batch_x = [x[batch] for x in self.multihop_feats]
        batch_x = torch.stack(batch_x)

        if self.label_emb is not None:
            batch_y_emb = self.label_emb[batch]
        else:
            batch_y_emb = None
        y = self.dataset.data.y[batch]
        return [batch_x, batch_y_emb, y]

    def test_transform(self, batch):
        return self.val_transform(batch)

    def pre_stage(self, stage, model_w_out):
        dataset = self.dataset
        probs = model_w_out
        with torch.no_grad():
            (label_emb, labels_with_pseudos, train_nid_with_pseudos) = prepare_labels(
                dataset, stage, self.label_nhop, self.threshold, probs=probs
            )
        self.label_emb = label_emb
        self.labels_with_pseudos = labels_with_pseudos
        self.train_nid_with_pseudos = train_nid_with_pseudos



import torch
import torch.nn as nn

from .. import ModelWrapper
from cogdl.utils.link_prediction_utils import cal_mrr, DistMultLayer, ConvELayer


class GNNKGLinkPredictionModelWrapper(ModelWrapper):
    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--score-func", type=str, default="distmult")
        # fmt: on

    def __init__(self, model, optimizer_cfg, score_func):
        super(GNNKGLinkPredictionModelWrapper, self).__init__()

        self.model = model
        self.optimizer_cfg = optimizer_cfg
        hidden_size = optimizer_cfg["hidden_size"]

        self.score_func = score_func
        if score_func == "distmult":
            self.scoring = DistMultLayer()
        elif score_func == "conve":
            self.scoring = ConvELayer(hidden_size)
        else:
            raise NotImplementedError

    def train_step(self, subgraph):
        graph = subgraph
        mask = graph.train_mask
        edge_index = torch.stack(graph.edge_index)
        edge_index, edge_types = edge_index[:, mask], graph.edge_attr[mask]

        with graph.local_graph():
            graph.edge_index = edge_index
            graph.edge_attr = edge_types
            loss = self.model.loss(graph, self.scoring)
        return loss

    def val_step(self, subgraph):
        train_mask = subgraph.train_mask
        eval_mask = subgraph.val_mask
        return self.eval_step(subgraph, train_mask, eval_mask)

    def test_step(self, subgraph):
        infer_mask = subgraph.train_mask | subgraph.val_mask
        eval_mask = subgraph.test_mask
        return self.eval_step(subgraph, infer_mask, eval_mask)

    def eval_step(self, graph, mask1, mask2):
        row, col = graph.edge_index
        edge_types = graph.edge_attr

        with graph.local_graph():
            graph.edge_index = (row[mask1], col[mask1])
            graph.edge_attr = edge_types[mask1]
            output, rel_weight = self.model.predict(graph)

        mrr, hits = cal_mrr(
            output,
            rel_weight,
            (row[mask2], col[mask2]),
            edge_types[mask2],
            scoring=self.scoring,
            protocol="raw",
            batch_size=500,
            hits=[1, 3, 10],
        )

        return dict(mrr=mrr, hits1=hits[0], hits3=hits[1], hits10=hits[2])

    def setup_optimizer(self):
        lr, weight_decay = self.optimizer_cfg["lr"], self.optimizer_cfg["weight_decay"]
        return torch.optim.AdamW(self.parameters(), lr=lr, weight_decay=weight_decay)

    def set_early_stopping(self):
        return "mrr", ">"



import json
import os

import torch
from cogdl.utils import download_url, untar
from transformers import BertTokenizer
import sentencepiece as spm

from .bert_model import BertConfig, BertForPreTrainingPreLN
from .oagbert_metainfo import OAGMetaInfoBertModel

PRETRAINED_MODEL_ARCHIVE_MAP = {
    "oagbert-v1": "https://cloud.tsinghua.edu.cn/f/051c9f87d8544698826e/?dl=1",
    "oagbert-test": "https://cloud.tsinghua.edu.cn/f/68a8d42802564d43984e/?dl=1",
    "oagbert-v2-test": "https://cloud.tsinghua.edu.cn/f/baff5abe84c4483bb690/?dl=1",
    "oagbert-v2": "https://cloud.tsinghua.edu.cn/f/f06448fa3c234317bd16/?dl=1",
    "oagbert-v2-lm": "https://cloud.tsinghua.edu.cn/f/efb2094951a94084947d/?dl=1",
    "oagbert-v2-sim": "https://cloud.tsinghua.edu.cn/f/103a467b0fe14177bf51/?dl=1",
    "oagbert-v2-zh": "https://cloud.tsinghua.edu.cn/f/8efc97a77af046dc8d24/?dl=1",
    "oagbert-v2-zh-sim": "https://cloud.tsinghua.edu.cn/f/6faaf0d095ee4013a098/?dl=1",
}


class OAGBertPretrainingModel(BertForPreTrainingPreLN):
    def __init__(self, bert_config):
        super(OAGBertPretrainingModel, self).__init__(bert_config)

    def forward(
        self,
        input_ids,
        token_type_ids=None,
        attention_mask=None,
        output_all_encoded_layers=False,
        checkpoint_activations=False,
    ):
        return self.bert.forward(
            input_ids=input_ids,
            token_type_ids=token_type_ids,
            attention_mask=attention_mask,
            output_all_encoded_layers=output_all_encoded_layers,
            checkpoint_activations=checkpoint_activations,
        )

    @staticmethod
    def _load(model_name_or_path: str, load_weights: bool = False):
        if not os.path.exists(model_name_or_path):
            if model_name_or_path in PRETRAINED_MODEL_ARCHIVE_MAP:
                if not os.path.exists(f"saved/{model_name_or_path}"):
                    archive_file = PRETRAINED_MODEL_ARCHIVE_MAP[model_name_or_path]
                    download_url(archive_file, "saved/", f"{model_name_or_path}.zip")
                    untar("saved/", f"{model_name_or_path}.zip")
                model_name_or_path = f"saved/{model_name_or_path}"
            else:
                raise KeyError("Cannot find the pretrained model {}".format(model_name_or_path))

        try:
            version = open(os.path.join(model_name_or_path, "version")).readline().strip()
        except Exception:
            version = None

        bert_config = BertConfig.from_dict(json.load(open(os.path.join(model_name_or_path, "bert_config.json"))))
        if os.path.exists(os.path.join(model_name_or_path, "vocab.txt")):
            tokenizer = BertTokenizer.from_pretrained(model_name_or_path)
        elif os.path.exists(os.path.join(model_name_or_path, "vocab.model")):
            tokenizer = spm.SentencePieceProcessor(model_file=os.path.join(model_name_or_path, "vocab.model"))
        else:
            raise FileNotFoundError("Cannot find vocabulary file")
        if version == "2":
            bert_model = OAGMetaInfoBertModel(bert_config, tokenizer)
        else:
            bert_model = OAGBertPretrainingModel(bert_config)

        model_weight_path = os.path.join(model_name_or_path, "pytorch_model.bin")
        if load_weights and os.path.exists(model_weight_path):
            bert_model.load_state_dict(torch.load(model_weight_path))

        return bert_config, tokenizer, bert_model


def oagbert(model_name_or_path="oagbert-v1", load_weights=True):
    """
    load oagbert model, return the underlying torch module and tokenizer. Tokenizer can be either BertTokenizer (en) or SentencePieceTokenizer (zh).
    """
    _, tokenizer, bert_model = OAGBertPretrainingModel._load(model_name_or_path, load_weights)

    return tokenizer, bert_model



from cogdl.wrappers import fetch_model_wrapper, fetch_data_wrapper


def set_default_wrapper_config():
    node_classification_models = [
        "gcn",
        "deepergcn",
        "drgcn",
        "drgat",
        "gcnii",
        "gcnmix",
        "grand",
        "grace",
        "mvgrl",
        "graphsage",
        "sage",
        "gdc_gcn",
        "mixhop",
        "mlp",
        "moe_gcn",
        "ppnp",
        "appnp",
        "pprgo",
        "chebyshev",
        "unet",
        "srgcn",
        "revgcn",
        "revgat",
        "revgen",
        "sagn",
        "sign",
        "sgc",
        "unsup_graphsage",
        "dgi",
        "dropedge_gcn",
        "gat",
        "graphsaint",
        "m3s",
        "correct_smooth_mlp",
    ]

    graph_classification_models = ["gin", "patchy_san", "diffpool", "infograph", "dgcnn", "sortpool"]

    network_embedding_models = [
        "deepwalk",
        "line",
        "node2vec",
        "prone",
        "netmf",
        "netsmf",
        "sdne",
        "spectral",
        "dngr",
        "grarep",
        "hope",
    ]

    graph_embedding_models = [
        "dgk",
        "graph2vec",
    ]

    graph_clustering_models = [
        "agc",
        "daegc",
        "gae",
        "vgae",
    ]

    graph_kg_link_prediction = ["rgcn", "compgcn"]

    node_classification_wrappers = dict()
    for item in node_classification_models:
        node_classification_wrappers[item] = {"mw": "node_classification_mw", "dw": "node_classification_dw"}

    node_classification_wrappers["dgi"]["mw"] = "dgi_mw"
    node_classification_wrappers["m3s"]["mw"] = "m3s_mw"
    node_classification_wrappers["graphsage"]["mw"] = "graphsage_mw"
    node_classification_wrappers["unsup_graphsage"]["mw"] = "unsup_graphsage_mw"
    node_classification_wrappers["mvgrl"]["mw"] = "mvgrl_mw"
    node_classification_wrappers["sagn"]["mw"] = "sagn_mw"
    node_classification_wrappers["grand"]["mw"] = "grand_mw"
    node_classification_wrappers["gcnmix"]["mw"] = "gcnmix_mw"
    node_classification_wrappers["grace"]["mw"] = "grace_mw"
    node_classification_wrappers["pprgo"]["mw"] = "pprgo_mw"

    node_classification_wrappers["m3s"]["dw"] = "m3s_dw"
    node_classification_wrappers["graphsage"]["dw"] = "graphsage_dw"
    node_classification_wrappers["pprgo"]["dw"] = "pprgo_dw"
    node_classification_wrappers["sagn"]["dw"] = "sagn_dw"

    graph_classification_wrappers = dict()
    for item in graph_classification_models:
        graph_classification_wrappers[item] = {"mw": "graph_classification_mw", "dw": "graph_classification_dw"}

    graph_classification_wrappers["infograph"] = {"mw": "infograph_mw", "dw": "infograph_dw"}

    network_embedding_wrappers = dict()
    for item in network_embedding_models:
        network_embedding_wrappers[item] = {"mw": "network_embedding_mw", "dw": "network_embedding_dw"}

    graph_embedding_wrappers = dict()
    for item in graph_embedding_models:
        graph_embedding_wrappers[item] = {"mw": "graph_embedding_mw", "dw": "graph_embedding_dw"}

    graph_clustering_wrappers = dict()
    for item in graph_clustering_models:
        graph_clustering_wrappers[item] = {"dw": "node_classification_dw"}
    graph_clustering_wrappers["gae"]["mw"] = "gae_mw"
    graph_clustering_wrappers["vgae"]["mw"] = "gae_mw"
    graph_clustering_wrappers["agc"]["mw"] = "agc_mw"
    graph_clustering_wrappers["daegc"]["mw"] = "daegc_mw"

    graph_kg_link_prediction_wrappers = dict()
    for item in graph_kg_link_prediction:
        graph_kg_link_prediction_wrappers[item] = {"dw": "gnn_kg_link_prediction_dw", "mw": "gnn_kg_link_prediction_mw"}

    other_wrappers = dict()
    other_wrappers["gatne"] = {"mw": "multiplex_embedding_mw", "dw": "multiplex_embedding_dw"}

    merged = dict()
    merged.update(node_classification_wrappers)
    merged.update(graph_embedding_wrappers)
    merged.update(graph_classification_wrappers)
    merged.update(network_embedding_wrappers)
    merged.update(graph_clustering_wrappers)
    merged.update(graph_kg_link_prediction_wrappers)
    merged.update(other_wrappers)
    return merged


default_wrapper_config = set_default_wrapper_config()


def get_wrappers(model_name):
    if model_name in default_wrapper_config:
        dw = default_wrapper_config[model_name]["dw"]
        mw = default_wrapper_config[model_name]["mw"]
        return fetch_model_wrapper(mw), fetch_data_wrapper(dw)


def get_wrappers_name(model_name):
    if model_name in default_wrapper_config:
        dw = default_wrapper_config[model_name]["dw"]
        mw = default_wrapper_config[model_name]["mw"]
        return mw, dw



from .. import BaseModel
from cogdl.layers import MLP as MLPLayer
from cogdl.data import Graph


class MLP(BaseModel):
    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--num-classes", type=int)
        parser.add_argument("--hidden-size", type=int, default=16)
        parser.add_argument("--num-layers", type=int, default=2)
        parser.add_argument("--dropout", type=float, default=0.5)
        parser.add_argument("--norm", type=str, default=None)
        parser.add_argument("--activation", type=str, default="relu")
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.num_classes,
            args.hidden_size,
            args.num_layers,
            args.dropout,
            args.activation,
            args.norm,
            args.act_first if hasattr(args, "act_first") else False,
        )

    def __init__(
        self,
        in_feats,
        out_feats,
        hidden_size,
        num_layers,
        dropout=0.0,
        activation="relu",
        norm=None,
        act_first=False,
        bias=True,
    ):
        super(MLP, self).__init__()
        self.nn = MLPLayer(in_feats, out_feats, hidden_size, num_layers, dropout, activation, norm, act_first, bias)

    def forward(self, x):
        if isinstance(x, Graph):
            x = x.x
        return self.nn(x)

    def predict(self, data):
        return self.forward(data.x)



from sklearn.metrics import roc_auc_score

import torch
from .. import ModelWrapper
from cogdl.utils import negative_edge_sampling


class GNNLinkPredictionModelWrapper(ModelWrapper):
    def __init__(self, model, optimizer_cfg):
        super(GNNLinkPredictionModelWrapper, self).__init__()
        self.model = model
        self.optimizer_cfg = optimizer_cfg
        self.loss_fn = torch.nn.BCELoss()

    def train_step(self, subgraph):
        graph = subgraph

        train_neg_edges = negative_edge_sampling(graph.train_edges, graph.num_nodes).to(self.device)
        train_pos_edges = graph.train_edges
        edge_index = torch.cat([train_pos_edges, train_neg_edges], dim=1)
        labels = self.get_link_labels(train_pos_edges.shape[1], train_neg_edges.shape[1], self.device)

        # link prediction loss
        with graph.local_graph():
            graph.edge_index = edge_index
            emb = self.model(graph)
        pred = (emb[edge_index[0]] * emb[edge_index[1]]).sum(1)
        pred = torch.sigmoid(pred)
        loss = self.loss_fn(pred, labels)
        return loss

    def val_step(self, subgraph):
        graph = subgraph
        pos_edges = graph.val_edges
        neg_edges = graph.val_neg_edges
        train_edges = graph.train_edges
        edges = torch.cat([pos_edges, neg_edges], dim=1)
        labels = self.get_link_labels(pos_edges.shape[1], neg_edges.shape[1], self.device).long()
        with graph.local_graph():
            graph.edge_index = train_edges
            with torch.no_grad():
                emb = self.model(graph)
                pred = (emb[edges[0]] * emb[edges[1]]).sum(-1)
        pred = torch.sigmoid(pred)

        auc_score = roc_auc_score(labels.cpu().numpy(), pred.cpu().numpy())

        self.note("auc", auc_score)

    def test_step(self, subgraph):
        graph = subgraph
        pos_edges = graph.test_edges
        neg_edges = graph.test_neg_edges
        train_edges = graph.train_edges
        edges = torch.cat([pos_edges, neg_edges], dim=1)
        labels = self.get_link_labels(pos_edges.shape[1], neg_edges.shape[1], self.device).long()
        with graph.local_graph():
            graph.edge_index = train_edges
            with torch.no_grad():
                emb = self.model(graph)
                pred = (emb[edges[0]] * emb[edges[1]]).sum(-1)
        pred = torch.sigmoid(pred)

        auc_score = roc_auc_score(labels.cpu().numpy(), pred.cpu().numpy())

        self.note("auc", auc_score)

    @staticmethod
    def get_link_labels(num_pos, num_neg, device=None):
        labels = torch.zeros(num_pos + num_neg)
        labels[:num_pos] = 1
        if device is not None:
            labels = labels.to(device)
        return labels.float()

    def setup_optimizer(self):
        lr, wd = self.optimizer_cfg["lr"], self.optimizer_cfg["weight_decay"]
        return torch.optim.Adam(self.parameters(), lr=lr, weight_decay=wd)

    def set_early_stopping(self):
        return "auc", ">"
